'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from _io import StringIO

from geopandas.geodataframe import GeoDataFrame
from pandas import read_csv
from shapely.geometry import Point
from shapely.wkt import loads
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.MyEdge import MyEdge
from t4gpd.commons.MyNode import MyNode


class GeoWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile, characteristicLength=10.0,
                 toLocalCrs=True, withPhysicalSurfaces=False):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.outputFile = outputFile
        self.lc = characteristicLength
        self.toLocalCrs = toLocalCrs
        self.withPhysicalSurfaces = withPhysicalSurfaces

    def run(self):
        newOrigin = Point((0.0, 0.0, 0.0))
        if self.toLocalCrs:
            minx, miny, _, _ = self.inputGdf.total_bounds
            newOrigin = Point((minx, miny, 0.0))

        mapOfNodes = {}
        mapOfEdges = {}
        arrayOfLineLoops = []
        arrayOfPlaneSurfaces = []
        mapOfPhysicalSurfaces = {}

        for _, row in self.inputGdf.iterrows():
            geom = row.geometry
            if GeomLib.isMultipart(geom):
                for g in geom.geoms:
                    self.__dumpPolygon(g, mapOfNodes, mapOfEdges, arrayOfLineLoops,
                                       arrayOfPlaneSurfaces, mapOfPhysicalSurfaces)
            else:
                self.__dumpPolygon(geom, mapOfNodes, mapOfEdges, arrayOfLineLoops,
                                   arrayOfPlaneSurfaces, mapOfPhysicalSurfaces)
 
        with open(self.outputFile, 'w') as output: 
            self.__dumpNodes(output, mapOfNodes, newOrigin)
            self.__dumpEdges(output, mapOfEdges, mapOfNodes)
            self.__dumpLineLoops(output, arrayOfLineLoops)
            self.__dumpPlaneSurfaces(output, arrayOfPlaneSurfaces)
            if self.withPhysicalSurfaces:
                self.__dumpPhysicalSurfaces(output, mapOfPhysicalSurfaces)

    def __dumpPolygon(self, polygon, mapOfNodes, mapOfEdges, arrayOfLineLoops,
                      arrayOfPlaneSurfaces, mapOfPhysicalSurfaces):
        tmp = list()
        tmp.append(self.__fulfillWith(mapOfNodes, mapOfEdges, arrayOfLineLoops, polygon.exterior))
        for hole in polygon.interiors:
            tmp.append(self.__fulfillWith(mapOfNodes, mapOfEdges, arrayOfLineLoops, hole))
        arrayOfPlaneSurfaces.append(tmp)

        if self.withPhysicalSurfaces:
            # Handle Physical Surfaces
            currPlaneSurfaceIdx = len(arrayOfPlaneSurfaces)
            _, _, nz = GeomLib3D.getFaceNormalVector(polygon)

            if Epsilon.isZero(nz):
                if 'Mur' in mapOfPhysicalSurfaces:
                    mapOfPhysicalSurfaces['Mur'].append(currPlaneSurfaceIdx)
                else:
                    mapOfPhysicalSurfaces['Mur'] = [currPlaneSurfaceIdx]
            elif Epsilon.equals(1.0, nz):
                if 'Toit' in mapOfPhysicalSurfaces:
                    mapOfPhysicalSurfaces['Toit'].append(currPlaneSurfaceIdx)
                else:
                    mapOfPhysicalSurfaces['Toit'] = [currPlaneSurfaceIdx]
            elif Epsilon.equals(-1.0, nz):
                if 'Sol' in mapOfPhysicalSurfaces:
                    mapOfPhysicalSurfaces['Sol'].append(currPlaneSurfaceIdx)
                else:
                    mapOfPhysicalSurfaces['Sol'] = [currPlaneSurfaceIdx]

    def __fulfillWith(self, mapOfNodes, mapOfEdges, arrayOfLineLoops, ring):
        nodes, tmp = list(ring.coords)[:-1], list()

        prev = MyNode(nodes[-1])
        for curr in nodes:
            curr = MyNode(curr)
            if curr not in mapOfNodes:
                mapOfNodes[curr] = 1 + len(mapOfNodes)

            myEdge = MyEdge(prev, curr)
            if myEdge not in mapOfEdges:
                mapOfEdges[myEdge] = 1 + len(mapOfEdges)
            myEdgeIdx = mapOfEdges[myEdge]

            if myEdge.startNode.__eq__(prev):
                tmp.append(myEdgeIdx)
            else:
                tmp.append(-myEdgeIdx)

            prev = curr

        arrayOfLineLoops.append(tmp)
        return len(arrayOfLineLoops)

    def __dumpCoords(self, node, newOrigin):
        return '%.2f, %.2f, %.2f' % (node.x - newOrigin.x, node.y - newOrigin.y, node.z - newOrigin.z)

    def __dumpNodes(self, output, mapOfNodes, newOrigin):
        output.write('lc = %g;\n' % self.lc)
        output.write('\n// =====> %d POINT(S)\n' % len(mapOfNodes))

        # for key, value in sorted(mapOfNodes.iteritems(), key=lambda(k, v): (v)):
        for key, value in sorted(list(mapOfNodes.items()), key=lambda kv: kv[1]):
            output.write('Point(%d) = {%s, lc};\n' % (value, self.__dumpCoords(key, newOrigin)))

    def __dumpEdges(self, output, mapOfEdges, mapOfNodes):
        output.write('\n// =====> %d LINE(S)\n' % len(mapOfEdges))
        # for key, value in sorted(mapOfEdges.iteritems(), key=lambda(k, v): v):
        for key, value in sorted(list(mapOfEdges.items()), key=lambda kv: kv[1]):
            output.write('Line(%d) = {%d, %d};\n' % (value, mapOfNodes[key.startNode], mapOfNodes[key.endNode]))

    def __dumpLineLoops(self, output, arrayOfLineLoops):
        output.write('\n// =====> %d LINE LOOP(S)\n' % len(arrayOfLineLoops))
        for i in range(len(arrayOfLineLoops)):
            output.write('Line Loop(%d) = {%s};\n' % (i + 1, ', '.join([str(x) for x in arrayOfLineLoops[i]])))

    def __dumpPlaneSurfaces(self, output, arrayOfPlaneSurfaces):
        output.write('\n// =====> %d PLANE SURFACE(S)\n' % len(arrayOfPlaneSurfaces))
        for i in range(len(arrayOfPlaneSurfaces)):
            output.write('Plane Surface(%d) = {%s};\n' % (i + 1, ', '.join([str(x) for x in arrayOfPlaneSurfaces[i]])))

    def __dumpPhysicalSurfaces(self, output, mapOfPhysicalSurfaces):
        output.write('\n// =====> %d PHYSICAL SURFACE(S)\n' % len(mapOfPhysicalSurfaces))
        for k, v in mapOfPhysicalSurfaces.items():
            output.write("Physical Surface('%s') = {%s};\n" % (
                k, ', '.join([str(x) for x in v])))



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from os import getpid, path
from subprocess import call
import tempfile

from geopandas.geodataframe import GeoDataFrame

from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.io.GeoWriter import GeoWriter
from t4gpd.io.MshReader import MshReader


class GmshTriangulator(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, characteristicLength=10.0, gmsh=None):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        self.characteristicLength = characteristicLength

        if gmsh is None:
            if path.exists('c:/Program Files (x86)/gmsh-2.11.0/gmsh.exe'):
                self.gmsh = 'c:/Program Files (x86)/gmsh-2.11.0/gmsh.exe'
            elif path.exists('/usr/local/bin/gmsh'):
                self.gmsh = '/usr/local/bin/gmsh'
            elif path.exists('/usr/bin/gmsh'):
                self.gmsh = '/usr/bin/gmsh'
            elif path.exists('/bin/gmsh'):
                self.gmsh = '/bin/gmsh'
            else:
                raise Exception('You must specify Gmsh location!')
        else:
            self.gmsh = gmsh

    def run(self):
        # sys.path.append('d:/bin/gmsh-2.11.0')
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpBasename = '_%d' % (getpid())
            geoTmpFile = '%s/%s.geo' % (tmpdir, tmpBasename)
            mshTmpFile = '%s/%s.msh' % (tmpdir, tmpBasename)

            GeoWriter(self.inputGdf, geoTmpFile, self.characteristicLength, True).run()
            print('GmshTriangulator:: write %s' % geoTmpFile)

            exitStatus = call([self.gmsh, '-2', geoTmpFile], stdin=None, stdout=None, stderr=None, shell=False)
            if 0 == exitStatus:
                print('GmshTriangulator:: gmsh -2 %s' % geoTmpFile)
    
                mshLayer = MshReader(mshTmpFile, self.inputGdf.total_bounds, self.inputGdf.crs).run()
                print('GmshTriangulator:: read %s' % mshTmpFile)
                return mshLayer



'''
Created on 31 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import ceil
from shapely.geometry import box
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.grid.AbstractGridLib import AbstractGridLib


class GridLib(AbstractGridLib):
    '''
    classdocs
    '''

    def __init__(self, gdf, dx, dy=None, encode=True):
        '''
        Constructor
        '''
        super().__init__(gdf, dx, dy, encode)

        self.minx, self.miny, maxx, maxy = gdf.total_bounds
        self.ncols = int(ceil((maxx - self.minx) / self.dx))
        self.nrows = int(ceil((maxy - self.miny) / self.dy))

        self.xOffset = ((self.ncols * self.dx) - (maxx - self.minx)) / 2.0
        self.yOffset = ((self.nrows * self.dy) - (maxy - self.miny)) / 2.0

    def __neighbors(self, r, c):
        e = r * self.ncols + (c + 1)
        ne = (r + 1) * self.ncols + (c + 1)
        n = (r + 1) * self.ncols + c
        nw = (r + 1) * self.ncols + (c - 1)
        w = r * self.ncols + (c - 1)
        sw = (r - 1) * self.ncols + (c - 1)
        s = (r - 1) * self.ncols + c
        se = (r - 1) * self.ncols + (c + 1)

        _na = -1
        if (0 == c):
            nw, w, sw = _na, _na, _na
        if (0 == r):
            sw, s, se = _na, _na, _na
        if (self.ncols == c + 1):
            ne, e, se = _na, _na, _na
        if (self.nrows == r + 1):
            nw, n, ne = _na, _na, _na
        return [e, n, w, s], [e, ne, n, nw, w, sw, s, se]

    def grid(self):
        result = []
        x0 = self.minx - self.xOffset
        for c in range(self.ncols):
            x0pp, y0 = x0 + self.dx, self.miny - self.yOffset
            for r in range(self.nrows):
                y0pp = y0 + self.dy
                neighbors4, neighbors8 = self.__neighbors(r, c)
                if self.encode:
                    neighbors4, neighbors8 = (ArrayCoding.encode(neighbors4),
                                              ArrayCoding.encode(neighbors8))
                result.append({
                    'gid': r * self.ncols + c,
                    'neighbors4': neighbors4,
                    'neighbors8': neighbors8,
                    'geometry': box(x0, y0, x0pp, y0pp)
                })
                y0 = y0pp
            x0 = x0pp

        return GeoDataFrame(result, crs=self.gdf.crs)



'''
Created on 7 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LineString

from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.commons.AngleLib import AngleLib


class CanyonStreetOrientation(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildings, magnitude=10.0):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings
        self.spatialIndex = buildings.sindex
        self.magnitude = magnitude

    def runWithArgs(self, row):
        vp = row.geometry.centroid

        (ux, uy), _ = RayCastingLib.prepareOrientedRays(self.buildings, self.spatialIndex, vp)

        p1 = (vp.x - self.magnitude * uy, vp.y + self.magnitude * ux)
        p2 = (vp.x + self.magnitude * uy, vp.y - self.magnitude * ux)

        azim = AngleLib.toDegrees(AngleLib.normAzimuth([-uy, ux]))
        azim = azim - (180 if (180 <= azim) else 0)

        return {
            'geometry': LineString((p1, p2)),
            'azimuth': azim
            }



'''
Created on 11 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas.core.frame import DataFrame
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from t4gpd.comfort.indices.AbstractThermalComfortIndice import AbstractThermalComfortIndice


class HI(AbstractThermalComfortIndice):
    '''
    classdocs
    '''

    def __init__(self, sensorsGdf, AirTC='AirTC_Avg', RH='RH_Avg'):
        '''
        Constructor

        AirTC: air temperature [C]
        RH: relative humidity [%]

        HI: Heat Index stated in Coccolo et al. (2016) [C]
        '''
        if not isinstance(sensorsGdf, DataFrame):
            raise IllegalArgumentTypeException(sensorsGdf, 'DataFrame')

        for fieldname in (AirTC, RH):
            if fieldname not in sensorsGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)

        self.AirTC = AirTC
        self.RH = RH

    def runWithArgs(self, row):
        AirTC = row[self.AirTC]
        RH = row[self.RH]

        # HI: Heat Index stated in Coccolo et al. (2016) [C]
        # (Grosdemouge 2020, pp. 109-110)
        HI = None

        if (20 < AirTC):
            HI = (-8.784695
                  +1.61139411 * AirTC
                  +2.338549 * RH
                  -0.14611605 * AirTC * RH 
                  -1.2308094e-2 * AirTC ** 2
                  -1.6424828e-2 * RH ** 2 
                  +2.211732e-3 * AirTC ** 2 * RH
                  +7.2546e-4 * AirTC * RH ** 2 
                  -3.582e-6 * AirTC ** 2 * RH ** 2)

        return { 'HI': HI }



'''
Created on 19 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import arcsin, arctan2, cos, sin, sqrt


class SphericalCartesianCoordinates(object):
    '''
    classdocs
    '''

    @staticmethod
    def fromCartesianToSphericalCoordinates(x, y, z):
        r = sqrt((x ** 2) + (y ** 2) + (z ** 2))
        if (0 == r):
            return 0, 0, 0
        lon = arctan2(y, x)  # LONGITUDE
        lat = arcsin(z / r)  # LATITUDE
        return r, lon, lat

    @staticmethod
    def fromSphericalToCartesianCoordinates(r, lon, lat):
        x = r * cos(lon) * cos(lat)
        y = r * sin(lon) * cos(lat)
        z = r * sin(lat)
        return x, y, z



'''
Created on 12 juin 2020

@author: tleduc

https://packaging.python.org/tutorials/packaging-projects/
'''
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

# with open('HISTORY.md') as history_file:
#     HISTORY = history_file.read()

# EXCLUSION_LIST = ['tests', 'tests.*']
EXCLUSION_LIST = ['tests', 'tests.*', 't4gpd.future', 't4gpd.future.*']

setup_args = dict(
    name='t4gpd',
    # version='0.0.1',
    version=open('t4gpd/_version.py').readlines()[-1].split()[-1].strip("\"'"),
    description='Set of tools based on Python, GeoPandas and Shapely to achieve urban geoprocessing',
    long_description_content_type='text/markdown',
    long_description=README + '\n\n',
    # long_description=README + '\n\n' + HISTORY,
    license='GPLv3+',
    packages=find_packages(exclude=EXCLUSION_LIST),
    author='Thomas Leduc',
    author_email='thomas.leduc@crenau.archi.fr',
    keywords=['Geospatial analysis', 'Urban form', 'Urban morphology', 'Isovist'],
    url='https://t4gpd-docs.readthedocs.io',
    download_url='https://sourcesup.renater.fr/frs/?group_id=463',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Education',
        'Topic :: Scientific/Engineering :: GIS',
    ],
    python_requires='>=3.6',
)

install_requires = [
    # 'Dijkstar',
    # 'geopandas',
    # 'matplotlib>=2.0.1',
    # 'networkx',
    # 'numpy',
    # 'opencv-python',
    # 'pysolar',
    # 'pythermalcomfort',
    # 'pyvista',
    # 'PyWavelets',
    # 'scikit-learn',
    # 'scipy',
    # 'shapely',
    # 'suntimes'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)



'''
Created on 5 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray, cos, ndarray, pi, sin
from numpy.linalg import norm
from pyvista.core.pointset import UnstructuredGrid
from shapely.geometry import LineString, Point
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from vtkmodules.vtkCommonCore import vtkIdList, vtkPoints
from vtkmodules.vtkFiltersGeneral import vtkOBBTree


class RayCastingIn3DLib(object):
    '''
    classdocs
    '''
    INFINITY = float('inf')

    @staticmethod
    def areCovisibleObbTree(obbTree, srcPt, dstPt, epsilon=1e-6):
        assert isinstance(obbTree, vtkOBBTree), 'obbTree must be a vtkOBBTree!'
        _points = vtkPoints()
        _cellIds = vtkIdList()
        code = obbTree.IntersectWithLine(srcPt, dstPt, _points, None)
        
        if (0 == code):
            # NO POINTS HIT
            return True
        else:
            pointData = _points.GetData()
            noPoints = pointData.GetNumberOfTuples()
            if (2 == noPoints):
                _distSrc = norm(asarray(pointData.GetTuple3(0)) - asarray(srcPt))
                _distDst = norm(asarray(pointData.GetTuple3(1)) - asarray(dstPt))
                if (_distSrc < epsilon) and (_distDst < epsilon):
                    return True
            elif (1 == noPoints):
                _distSrc = norm(asarray(pointData.GetTuple3(0)) - asarray(srcPt))
                _distDst = norm(asarray(pointData.GetTuple3(0)) - asarray(dstPt))
                if (_distSrc < epsilon) or (_distDst < epsilon):
                    return True
        return False

    @staticmethod
    def mraycastObbTree(obbTree, srcPts, dstPts, pkFieldname=None):
        '''
        Excerpt from https://www.kitware.com/ray-casting-ray-tracing-with-vtk/

        the vtkOBBTree class, which generates an oriented bounding-box (OBB) ‘tree’
        for a given geometry under a vtkPolyData object. Upon generation of this OBB 
        tree, the vtkOBBTree class allows us to perform intersection tests between
        the mesh and the lines of finite length, as well as intersection tests between
        different meshes. It can then return the point coordinates where intersections
        were detected, as well as the polydata cell IDs where the intersections occurred.
        '''
        assert isinstance(obbTree, vtkOBBTree), 'obbTree must be a vtkOBBTree!'
        assert len(srcPts) == len(dstPts), 'The number of origin points must be equal to the number of target points!'

        rays, hitPoints, hitDists, hitGids = [], [], [], []
        for i, srcPt in enumerate(srcPts):
            dstPt = dstPts[i]
            _points = vtkPoints()
            _cellIds = vtkIdList()
            code = obbTree.IntersectWithLine(srcPt, dstPt, _points, _cellIds)
            if (0 == code):
                # NO POINTS HIT
                ray, hitPoint, hitDist, hitGid = [srcPt, dstPt], dstPt, RayCastingIn3DLib.INFINITY, None
            else:
                pointData = _points.GetData()
                noPoints = pointData.GetNumberOfTuples()
                hitDist = RayCastingIn3DLib.INFINITY
                hitGid = None
                for idx in range(noPoints):
                    _tup = pointData.GetTuple3(idx)
                    _dist = norm(asarray(_tup) - asarray(srcPt))
                    if (_dist < hitDist):
                        ray, hitPoint, hitDist = [srcPt, _tup], _tup, _dist
                        if (0 < idx):
                            print('Surprise in RayCastingIn3DLib.raycast(...)!')
                        if pkFieldname is not None:
                            hitGid = _cellIds.GetId(idx)

            rays.append(ray)
            hitPoints.append(hitPoint)
            hitDists.append(hitDist)
            hitGids.append(hitGid)
        return rays, hitPoints, hitDists, hitGids

    @staticmethod
    def mraycast(unstructuredGrid, srcPts, dstPts, pkFieldname=None):
        obbTree = RayCastingIn3DLib.prepareVtkOBBTree(unstructuredGrid)
        return RayCastingIn3DLib.mraycastObbTree(obbTree, srcPts, dstPts, pkFieldname)

    @staticmethod
    def mraycast2D(unstructuredGrid, viewpoint, nRays=36):
        if isinstance(viewpoint, Point):
            viewpoint = viewpoint.coords[0]

        angularOffset, raylen = (2.0 * pi) / nRays, unstructuredGrid.length
        srcPts = [viewpoint] * nRays
        dstPts = [[
            viewpoint[0] + raylen * cos(angularOffset * i),
            viewpoint[1] + raylen * sin(angularOffset * i),
            0 if (2 == len(viewpoint)) else viewpoint[2],
            ] for i in range(nRays)]
        return RayCastingIn3DLib.mraycast(unstructuredGrid, srcPts, dstPts)

    @staticmethod
    def mraycast2DToGeoDataFrame(masks, viewpoint, nRays=36, crs='epsg:2154'):
        if isinstance(masks, (list, tuple, ndarray)):
            uGrid = ToUnstructuredGrid(masks).run()
        elif isinstance(masks, UnstructuredGrid):
            uGrid = masks
        else:
            raise IllegalArgumentTypeException(masks, 'UnstructuredGrid')
        rays, hitPoints, hitDists, hitGids = RayCastingIn3DLib.mraycast2D(uGrid, viewpoint, nRays)
        rows = []
        for ray, hitPoint, hitDist in zip(rays, hitPoints, hitDists):
            rows.append({
                'geometry': LineString(ray),
                'hitPoint': Point(hitPoint),
                'hitDist': hitDist
                })
        return GeoDataFrame(rows, crs=crs)

    @staticmethod
    def prepareVtkOBBTree(*args):
        '''
        Excerpt from https://www.kitware.com/ray-casting-ray-tracing-with-vtk/

        the vtkOBBTree class, which generates an oriented bounding-box (OBB) 'tree'
        for a given geometry under a vtkPolyData object. Upon generation of this OBB 
        tree, the vtkOBBTree class allows us to perform intersection tests between
        the mesh and the lines of finite length, as well as intersection tests between
        different meshes. It can then return the point coordinates where intersections
        were detected, as well as the polydata cell IDs where the intersections occurred.
        '''
        if all([isinstance(args[i], GeoDataFrame) for i in range(len(args))]):
            uGrid = ToUnstructuredGrid(args).run()
        elif (1 == len(args)) and isinstance(args[0], UnstructuredGrid):
            uGrid = args[0]
        else:
            raise IllegalArgumentTypeException(args, 'single UnstructuredGrid or list of GeoDataFrame')

        obbTree = vtkOBBTree()
        obbTree.SetDataSet(uGrid.extract_geometry())
        obbTree.BuildLocator()

        return obbTree



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import abs, exp, mean, sin, sqrt

from t4gpd.comfort.algo.ConstantsLib import ConstantsLib


class CommonsLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def airPressure(AirTC, RH):
        # Air pressure in Pascal and mPa
        # Saturated water pressure
        P_s = exp(18.956 - (4030.183 / (AirTC + 235)))  # [mb] 10mb=1kPa=1000Pa=1000N.m-2
        # Partial air pressure at T_air in Pascal
        P_a = RH * P_s
        # Partial air pressure at T_air in [kPa]
        P_air = P_a / 1000
        return P_a, P_air

    @staticmethod
    def convectiveHeatTransferCoefficient(AirTC, WS_ms, P_a):
        # calculation of hc parameter depending on wind speed
        if (0.1 <= WS_ms):
            # convective heat transfer coefficient (W m2 K???1) after Staiger et al. (2012) as 
            # for external conditions only forced convection v_air > 0.1 m/s 
            hc = 12.1 * sqrt(WS_ms * P_a / 1013.25)
        else:
            # convective heat transfer coefficient (W m???2 K???1) after Parsons (2003) and Da Silva
            var = WS_ms + 0.0052 * (ConstantsLib.M - 58)
            hc_1 = 2.38 * abs(ConstantsLib.tsk - AirTC) ** 0.25  
            hc_2 = 3.5 + 5.2 * var
            hc_3 = 8.7 * var ** 0.6
            # max value of hc between natural and forced convection
            hc = max(hc_1, hc_2, hc_3)
        return hc

    @staticmethod
    def iterationEtu(w, heo, P_air, T_ao, hu, NUATF, SERFL, ERFS, EVCF):
        '''
        function to approximate ETU after Watanabe et al. (2014)
        function for iterating the the universal effective temperature in degrees Celsius (C)
        '''
        # P_ETUs saturated water vapor pressure at ETU, kPa
        P_ETUs = exp(18.956 - (4030.183 / (T_ao + 235)))  # [mb] #first assumption that ETU is T_ao?????
        P_ETUs = P_ETUs / 10  # [kPa]
        
        # standard effective humid field
        SEHF = w * heo * (P_air - 0.5 * P_ETUs)
        
        ETU = T_ao + 1 / hu * (NUATF + SERFL + ERFS + SEHF + EVCF)
  
        while True:
            err_ETU = ETU - (T_ao + 1 / hu * (NUATF + SERFL + ERFS + SEHF + EVCF))
            if (0.001 < abs(err_ETU)):
                ETU = ETU + 0.1
                P_ETUs = exp(18.956 - (4030.183 / (ETU + 235)))  # [mb]
                P_ETUs = P_ETUs / 10  # [kPa]
                SEHF = w * heo * (P_air - 0.5 * P_ETUs)
            else:
                return ETU

    @staticmethod
    def iterationTcl(tsk, Icl, fcl, hc, T_kel, T_air, T_mrt, tcl_init):
        '''
        # ITERATION_TCL
        # for PMV  and PT calculation
        # function for iterating the clothing surface temperature, in degrees Celsius (C)
        '''
        oldTcl = tcl_init
        dif_min_tcl = 0.001  # C

        while True:
            newTcl = (oldTcl - 
                      ((oldTcl - tsk + Icl * fcl * 
                        (3.96e-8 * ((oldTcl + T_kel) ** 4 - (T_mrt + T_kel) ** 4))
                        +hc * (oldTcl - T_air)) / 
                        (1 + Icl * fcl * (3.96e-8 * 4 * (oldTcl + T_kel) ** 3
                                          +hc * (oldTcl)))))  # after Staiger et al. (2012)

            dif = abs(oldTcl - newTcl)
            oldTcl = mean([newTcl, oldTcl])
            if (dif < dif_min_tcl):
                return oldTcl
            oldTcl = newTcl

    @staticmethod
    def radiantHeatTransferCoefficient(AirTC, RH, WS_ms, T_mrt):
        # frd effective radiation area factor [-] assumed to be alpha_eff 
        frd = ConstantsLib.alpha_eff

        P_a, _ = CommonsLib.airPressure(AirTC, RH)

        hc = CommonsLib.convectiveHeatTransferCoefficient(AirTC, WS_ms, P_a)

        # initial clothing surface temperature, in degrees Celsius (C)
        tcl_init = AirTC

        # Call function "ITERATION_TCL" the clothing surface temperature, in degrees Celsius (C)
        tcl = CommonsLib.iterationTcl(
            ConstantsLib.tsk, ConstantsLib.Icl, ConstantsLib.fcl, hc,
            ConstantsLib.T_kel, AirTC, T_mrt, tcl_init)

        # radiant heat transfer coefficient 
        hr = (ConstantsLib.epsi_p * ConstantsLib.sigma_B * frd * 
              ((tcl + ConstantsLib.T_kel) ** 4 - 
               (T_mrt + ConstantsLib.T_kel) ** 4) / 
              ((tcl + ConstantsLib.T_kel) - (T_mrt + ConstantsLib.T_kel)))
        return hr

    @staticmethod
    def SDIF_Idn(N, SR01Up_1):
        # Calculation of direct and diffuse shortwave radiation based on nebulosity approximation 
        # from MF weather station

        # Calculation of clearness index K and diffuse solar radiation coeff DIFF
        # K calcule en fonction de la regression quadratique de Black (1956)
        # DIFF calcule en fonction de Muneer et al. (2004)
        K = 0.803 - 0.458 * (N / 8.) ** 2 - 0.34 * (N / 8.)
        if (K < 0.2):
            DIFF = 0.98
        else:
            DIFF = 0.962 + 0.779 * K - 4.375 * K ** 2 + 2.716 * K ** 3

        # Calculation of direct and diffuse repartition of incoming global radiation (SWin=UP) 
        SDIR = (1.0 - DIFF) * SR01Up_1
        SDIF = DIFF * SR01Up_1

        # Idn [W.m-2] normal direct solar radiation
        Idn = SDIR / sin(ConstantsLib.h_sun_rad)

        return (SDIF, Idn)



'''
Created on 15 mars 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from matplotlib.colors import ListedColormap
from numpy import ndarray
from shapely.affinity import translate, scale
from shapely.geometry import Point, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STCompass(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, basemap, bounds, crs, magnitude=0.015):
        '''
        Constructor
        '''
        if isinstance(bounds, (list, tuple, ndarray)) and (4 == len(bounds)):
            self.basemap = basemap
            self.bounds = bounds
            self.crs = crs
            self.magnitude = magnitude
        else:
            raise IllegalArgumentTypeException(bounds, 'box (minx, miny, maxx, maxy)')

    @staticmethod
    def __letter():
        h, w = 1.45, 0.45
        return Polygon([(-0.5 - w, 3), (-0.5, 3), (-0.5, 3 + h - w), (0.5, 3),
                        (0.5 + w, 3), (0.5 + w, 3 + h), (0.5, 3 + h), (0.5, 3 + w),
                        (-0.5, 3 + h), (-0.5 - w, 3 + h)])

    @staticmethod
    def __left_arrow():
        return Polygon([(0, 0), (0, 2), (-2, -1.5)])

    @staticmethod
    def __right_arrow():
        return Polygon([(0, 0), (2, -1.5), (0, 2)])

    def __translation_and_dilation(self, arrow):
        minx, miny, maxx, maxy = self.bounds

        deltax, deltay = maxx - minx, maxy - miny 
        delta = min(deltax, deltay)

        arrow = scale(arrow, xfact=self.magnitude * delta, yfact=self.magnitude * delta, origin=Point([0, 0]))
        arrow = translate(arrow, xoff=minx + 2 * self.magnitude * delta, yoff=maxy - 4.5 * self.magnitude * delta)
        return arrow

    def run(self):
        _left_arrow = self.__translation_and_dilation(self.__left_arrow())
        _right_arrow = self.__translation_and_dilation(self.__right_arrow())
        _letter = self.__translation_and_dilation(self.__letter())

        my_rgb_cmap = ListedColormap(['white', 'black', 'black'])

        gdf = GeoDataFrame([
            {'gid': 0, 'geometry': _left_arrow},
            {'gid': 1, 'geometry': _right_arrow},
            {'gid': 2, 'geometry': _letter}
            ], crs=self.crs)
        gdf.plot(ax=self.basemap, edgecolor='black', column='gid', cmap=my_rgb_cmap)

        return gdf



'''
Created on 19 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from itertools import combinations
from shapely.ops import unary_union

from geopandas.geodataframe import GeoDataFrame

from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.ListUtilities import ListUtilities


class STMultipleOverlaps2(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, by=None):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        if (by is not None) and (by not in inputGdf):
            raise Exception('%s is not a relevant field name!' % (by))
        self.by = by

    def run(self):
        rows = []

        if self.by is None:
            i, allIntersections = 0, dict({ 1: dict() })
            for geom in self.inputGdf.geometry:
                _geoms = []
                for _geom in GeomLib.flattenGeometry(geom):
                    if 1.0 < _geom.area:
                        _geoms.append(_geom)
                if 0 < len(_geoms):
                    allIntersections[1][tuple([i])] = unary_union(_geoms)
                    i += 1

            nGeoms = len(allIntersections[1])
            geomIds = list(range(nGeoms))

            for n in range(2, nGeoms + 1):
                allIntersections[n] = dict()
                for combin in combinations(geomIds, n):
                    keyLeft = combin[:-1]
                    keyRight = tuple([combin[-1]])

                    allIntersections[n][combin] = allIntersections[n - 1][keyLeft].intersection(
                        allIntersections[1][keyRight])

            for n in sorted(allIntersections.keys()):
                row = { 'geometry': unary_union(allIntersections[n].values()), 'nOverlap': n }
                rows.append(row)

            for n in range(len(rows) - 1):
                rows[n]['geometry'] = rows[n]['geometry'].difference(rows[n + 1]['geometry'])

        else:
            allIntersections = dict({ 1: dict() })
            for _, row in self.inputGdf.iterrows():
                key, geom = row[self.by], row.geometry
                _geoms = []
                for _geom in GeomLib.flattenGeometry(geom):
                    if 1.0 < _geom.area:
                        _geoms.append(_geom)
                if 0 < len(_geoms):
                    allIntersections[1][tuple([key])] = unary_union(_geoms)

            nGeoms = len(allIntersections[1])
            geomIds = ListUtilities.flatten(allIntersections[1].keys())

            for n in range(2, nGeoms + 1):
                allIntersections[n] = dict()
                for combin in combinations(geomIds, n):
                    keyLeft = combin[:-1]
                    keyRight = tuple([combin[-1]])

                    allIntersections[n][combin] = allIntersections[n - 1][keyLeft].intersection(
                        allIntersections[1][keyRight])

            uog = None
            for n in range(nGeoms, 0, -1):
                for combin in combinations(geomIds, n):
                    _geom = allIntersections[n][combin]
                    if uog is not None:
                        _geom = _geom.difference(uog)

                    if 0.0 < _geom.area:
                        row = { 'geometry': _geom, 'nOverlap': n,
                               'matched_id': ArrayCoding.encode(combin) }
                        rows.append(row)
                uog = unary_union(allIntersections[n].values())

        return GeoDataFrame(rows, crs=self.inputGdf.crs)



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point, Polygon
from t4gpd.commons.GeoProcess import GeoProcess


class MshReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, bbox, crs='EPSG:2154'):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.crs = crs
        minx, miny, _, _ = bbox
        self.newOrigin = Point((minx, miny))

    def __getCoordinates(self, nodes, i1, i2, i3):
        # TL 10.03.2021: add z-components
        return Polygon([
            [nodes[i1].x + self.newOrigin.x, nodes[i1].y + self.newOrigin.y, nodes[i1].z],
            [nodes[i2].x + self.newOrigin.x, nodes[i2].y + self.newOrigin.y, nodes[i2].z],
            [nodes[i3].x + self.newOrigin.x, nodes[i3].y + self.newOrigin.y, nodes[i3].z]
            ])

    def __error(self, nline, line, expected):
        separator = '*' * 80
        msg = '''
%s
\tMshReader:: File format error (line %d):
\tActual: %s
\t=> Expecting: %s
%s
''' % (separator, nline, line, expected, separator)
        raise Exception(msg)

    def __getVersion(self):
        with open(self.inputFile, 'r') as f:
            for nline, line in enumerate(f, start=1):
                line = line.strip()
                if (1 == nline):
                    if not '$MeshFormat' == line: 
                        self.__error(nline, line, '$MeshFormat')
                elif (2 == nline):
                    tmp = line.split()
                    if 3 != len(tmp): 
                        self.__error(nline, line, 'version-number file-type data-size')
                    else:
                        # version_number, file_type, data_size = tmp
                        return int(float(tmp[0]))

    def __readV2(self):
        items = []
        with open(self.inputFile, 'r') as f:
            nbNodes, nbElements, nodes = 0, 0, dict()

            for nline, line in enumerate(f, start=1):
                line = line.strip()

                if (nline <= 3):
                    # READ SECTION "MeshFormat"
                    if (1 == nline):
                        if not '$MeshFormat' == line:
                            self.__error(nline, line, '$MeshFormat')
                    elif (2 == nline):
                        tmp = line.split()
                        if 3 != len(tmp): 
                            self.__error(nline, line, 'version-number file-type data-size')
                    elif (3 == nline):
                        if not '$EndMeshFormat' == line: 
                            self.__error(nline, line, '$EndMeshFormat')

                elif (nline <= 6 + nbNodes):
                    # READ SECTION "Nodes"
                    if (4 == nline):
                        if not '$Nodes' == line: 
                            self.__error(nline, line, '$Nodes')
                    elif (5 == nline):
                        nbNodes = int(line)
                    elif (nline <= 5 + nbNodes):
                        # READ ALL NODES
                        nodeNumber, x, y, z = line.split()
                        nodes[int(nodeNumber)] = Point((float(x), float(y), float(z)))
                        if int(nodeNumber) != len(nodes):
                            self.__error(nline, line, 'node-number x-coord y-coord z-coord')
                    else:
                        if not '$EndNodes' == line: 
                            self.__error(nline, line, '$EndNodes')

                else:
                    # READ SECTION "Elements"
                    if (7 + nbNodes == nline):
                        if not '$Elements' == line: 
                            self.__error(nline, line, '$Elements')
                    elif (8 + nbNodes == nline):
                        nbElements = int(line) 
                    elif (nline <= 8 + nbNodes + nbElements):
                        # READ ALL ELEMENTS
                        tmp = line.split()
                        elementNumber, elementType = [int(i) for i in tmp[:2]]
                        if (2 == elementType):
                            # 3-node triangle
                            i1, i2, i3 = [int(i) for i in tmp[-3:]]
                            items.append({'gid': int(elementNumber), 'geometry':self.__getCoordinates(nodes, i1, i2, i3)})
                    else:
                        if not '$EndElements' == line: 
                            self.__error(nline, line, '$Elements')
        '''
        nodes = [{'gid':k, 'geometry':Point((v.x + self.newOrigin.x, v.y + self.newOrigin.y))}
                 for k, v in nodes.items()]
        GeoDataFrame(nodes, crs=self.crs).to_file('/tmp/yyy.shp')
        '''
        return items

    def __analyzeSectionNodesV4(self, content):
        nodes, nline = dict(), 0
        while nline < len(content):
            # entityDim, entityTag, parametric, numNodesInBlock = content[nline].split()        
            entityDim, _, _, numNodesInBlock = content[nline].split()
            entityDim, numNodesInBlock = int(entityDim), int(numNodesInBlock)

            if (0 == entityDim):
                nodeTag = int(content[nline + 1])
                x, y, z = content[nline + 2].split()
                nodes[nodeTag] = Point((float(x), float(y), float(z)))
                nline += 3

            elif (entityDim in [1, 2]):
                for i in range(numNodesInBlock):
                    nodeTag = int(content[nline + 1 + i])
                    x, y, z = content[nline + 1 + numNodesInBlock + i].split()
                    nodes[nodeTag] = Point((float(x), float(y), float(z)))
                nline += 1 + 2 * numNodesInBlock
        return nodes

    def __analyzeSectionElementsV4(self, content, nodes):
        triangles, nline = dict(), 0
        while nline < len(content):
            # entityDim, entityTag, elementType, numElementsInBlock = content[nline].split()
            _, _, elementType, numElementsInBlock = content[nline].split()
            elementType, numElementsInBlock = int(elementType), int(numElementsInBlock)

            if (2 == elementType):
                # 3-node triangle
                for i in range(numElementsInBlock):
                    elementTag, nodeTag1, nodeTag2, nodeTag3 = [int(n) for n in content[nline + 1 + i].split()]
                    triangles[elementTag] = self.__getCoordinates(nodes, nodeTag1, nodeTag2, nodeTag3)
            nline += 1 + numElementsInBlock
        return triangles

    def __readV4(self):
        with open(self.inputFile, 'r') as f:
            nbPoints, nbCurves, nbSurfaces, nbVolumes, nbEntities = 0, 0, 0, 0, 0
            nodes = []

            contentOfSectionNodes, contentOfSectionElements = [], []

            for nline, line in enumerate(f, start=1):
                line = line.strip()

                if (nline <= 3):
                    # READ SECTION "MeshFormat"
                    if (1 == nline):
                        if not '$MeshFormat' == line:
                            self.__error(nline, line, '$MeshFormat')
                    elif (2 == nline):
                        tmp = line.split()
                        if 3 != len(tmp): 
                            self.__error(nline, line, 'version-number file-type data-size')
                    elif (3 == nline):
                        if not '$EndMeshFormat' == line: 
                            self.__error(nline, line, '$EndMeshFormat')

                elif (nline <= 6 + nbEntities):
                    # READ SECTION "Entities"
                    if (4 == nline):
                        if not '$Entities' == line: 
                            self.__error(nline, line, '$Entities')
                    elif (5 == nline):
                        nbPoints, nbCurves, nbSurfaces, nbVolumes = [int(n) for n in line.split()]
                        nbEntities = nbPoints + nbCurves + nbSurfaces + nbVolumes
                    elif (nline <= 4 + nbPoints + 1):
                        '''
                        # USELESS
                        pointTag, x, y, z, _ = line.split()
                        nodes.append(Point((float(x), float(y), float(z))))
                        if int(pointTag) != len(nodes):
                            self.__error(nline, line, 'pointTag X Y Z numPhysicalTags')
                        '''
                        pass
                    elif (6 + nbEntities == nline):
                        beginOfSectionNodes = True
                        if not '$EndEntities' == line: 
                            self.__error(nline, line, '$EndEntities')

                elif beginOfSectionNodes:
                    # READ SECTION "Nodes"
                    if (7 + nbEntities == nline):
                        if not '$Nodes' == line: 
                            self.__error(nline, line, '$Nodes')
                    elif (8 + nbEntities == nline):
                        # USELESS
                        # numEntityBlocks, numNodes, minNodeTag, maxNodeTag = [int(n) for n in line.split()]
                        pass                    
                    else:
                        if ('$' != line[0]):
                            contentOfSectionNodes.append(line)
                        else:
                            if not '$EndNodes' == line: 
                                self.__error(nline, line, '$EndNodes')

                            nodes = self.__analyzeSectionNodesV4(contentOfSectionNodes)
                            beginOfSectionNodes = False
                            beginOfSectionElements = True

                elif beginOfSectionElements:
                    # READ SECTION "Elements"
                    if (10 + nbEntities + len(contentOfSectionNodes) == nline):
                        if not '$Elements' == line: 
                            self.__error(nline, line, '$Elements')
                    elif (11 + nbEntities + len(contentOfSectionNodes) == nline):
                        # USELESS
                        # numEntityBlocks, numElements, minElementTag, maxElementTag = [int(n) for n in line.split()]
                        pass
                    else:
                        if ('$' != line[0]):
                            contentOfSectionElements.append(line)
                        else:
                            if not '$EndElements' == line: 
                                self.__error(nline, line, '$EndElements')

                            triangles = self.__analyzeSectionElementsV4(contentOfSectionElements, nodes)
                            beginOfSectionElements = False

        items = [{'geometry': ring, 'gid': gid} for gid, ring in triangles.items()]

        '''
        nodes = [{'gid':k, 'geometry':Point((v.x + self.newOrigin.x, v.y + self.newOrigin.y))}
                 for k, v in nodes.items()]
        GeoDataFrame(nodes, crs=self.crs).to_file('/tmp/yyy.shp')
        '''

        return items

    def run(self):
        version = self.__getVersion()
        if 2 == version:
            print('MshReader:: read V2')
            rows = self.__readV2()
        elif 4 == version:
            print('MshReader:: read V4')
            rows = self.__readV4()
        else: 
            raise NotImplementedError(f'\n\n\t*** MshReader:: file format version #{version} must be implemented! ***')

        return GeoDataFrame(rows, crs=self.crs)



'''
Created on 8 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import re

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Polygon
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.CSVLib import CSVLib
from t4gpd.io.AbstractReader import AbstractReader


class CirReader(AbstractReader):
    '''
    classdocs
    '''
    RE1 = re.compile(r'^f(\d+)$')
    RE2 = re.compile(r'^c(\d+)$')
    
    def __init__(self, inputFile):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.data = self.__initialParsing()
        self.ptr = 0

    def __initialParsing(self):
        result = []

        with CirReader.opener(self.inputFile) as f:
            for line in f:
                result += [CSVLib.readLexeme(value) for value in line.split()]

        return result

    def __read(self):
        self.ptr += 1
        return self.data[self.ptr - 1]
 
    def __read3Coords(self):
        return [self.__read(), self.__read(), self.__read()]

    def __readBBox(self):
        self.ptr += 10

    def __readContour(self):
        nbHoles = self.__readNumberOfHoles()
        extRing = self.__readPolyline()
        intRings = []
        for _ in range(nbHoles):
            if 't' == self.__read():
                intRings.append(self.__readPolyline())
        if 0 < nbHoles:
            return Polygon(extRing, intRings)
        return Polygon(extRing)

    def __readFace(self):
        faceNumber = self.__readFaceNumber()
        nbContours = self.__read()
        normal_vec = self.__read3Coords()
        result = []
        for ctrNumber in range(nbContours):
            gid = ArrayCoding.encode([faceNumber] if (1 == nbContours) else [faceNumber, ctrNumber])
            result.append({
                'cir_id': gid,
                'geometry': self.__readContour(),
                # 'normal_vec': ArrayCoding.encode(normal_vec) if self.encode else normal_vec
                })
        return result

    def __readFaceNumber(self):
        word = self.__read()
        return int(self.RE1.search(word).group(1))

    def __readNumberOfHoles(self):
        word = self.__read()
        return int(self.RE2.search(word).group(1))

    def __readPolyline(self):
        nbOfNodes, result = self.__read(), []
        for _ in range(nbOfNodes):
            result.append(self.__read3Coords())
        return result
        # return result[:-1]

    def run(self):
        result = []

        if 30 < len(self.data):
            nbFaces = self.__read()
            supNumFaces = self.__read()
            self.__readBBox()
            for _ in range(nbFaces):
                result += self.__readFace()

        return GeoDataFrame(result)



'''
Created on 21 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import copy

from dijkstar.graph import Graph
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, MultiLineString
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class AbstractUrbanGraphLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def addCoupleOfVertices(graph, edges, first, last, linestring):
        linestringLength = linestring.length

        graph.add_edge(first, last, (linestringLength, linestring.wkt))

        edgeItem = { 'geometry': linestring, 'length': linestringLength }
        if first in edges:
            if last in edges[first]:
                raise Exception('Unreachable instruction!')
            else:
                edges[first][last] = edgeItem
        else:
            edges[first] = dict({ last: edgeItem })

    @staticmethod
    def addEdge(graph, ciVertices, icVertices, edges, linestring):
        coords = linestring.coords
        if (2 <= len(coords)):
            first = AbstractUrbanGraphLib.addVertex(ciVertices, icVertices, coords[0])
            last = AbstractUrbanGraphLib.addVertex(ciVertices, icVertices, coords[-1])

            if (first in edges) and (first == last):
                print('This is a multigraph (loop)!')
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges1 = GeomLib.cutsLineStringByCurvilinearDistance(linestring, linestring.length / 3.0)
                newEdges2 = GeomLib.cutsLineStringByCurvilinearDistance(linestring, 2.0 * linestring.length / 3.0)
                newEdges = newEdges1 + newEdges2[:-1]
                for _edge in newEdges:
                    AbstractUrbanGraphLib.addEdge(graph, ciVertices, icVertices, edges, _edge)

            elif (first in edges) and (last in edges[first]):
                print('This is a multigraph!')
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges = GeomLib.cutsLineStringByCurvilinearDistance(linestring, linestring.length / 2.0)
                for _edge in newEdges:
                    AbstractUrbanGraphLib.addEdge(graph, ciVertices, icVertices, edges, _edge)

            else:
                AbstractUrbanGraphLib.addCoupleOfVertices(graph, edges, first, last, linestring)
                AbstractUrbanGraphLib.addCoupleOfVertices(graph, edges, last, first, LineString(reversed(coords)))

    @staticmethod
    def addVertex(ciVertices, icVertices, coord):
        str_coord = AbstractUrbanGraphLib.hashCoord(coord)

        if not str_coord in ciVertices:
            nodeIndex = len(ciVertices)
            ciVertices[str_coord] = nodeIndex
            icVertices[nodeIndex] = coord
        else:
            nodeIndex = ciVertices[str_coord]

        return nodeIndex   

    @staticmethod
    def buildGraph(geometries):
        graph = Graph()
        ciVertices = dict()
        icVertices = dict()
        edges = dict()

        for geom in geometries:
            if isinstance(geom, MultiLineString):
                for _geom in geom.geoms:
                    AbstractUrbanGraphLib.addEdge(graph, ciVertices, icVertices, edges, _geom)
            elif isinstance(geom, LineString):
                AbstractUrbanGraphLib.addEdge(graph, ciVertices, icVertices, edges, geom)
            else:
                raise IllegalArgumentTypeException(geom, 'LineString or MultiLineString')

        return graph, ciVertices, icVertices, edges

    @staticmethod
    def cost_function(u, v, edge, prev_edge):
        length, _ = edge
        return length

    @staticmethod
    def hashCoord(coord):
        str_coord = ('%f_%f') % (coord[0], coord[1])
        return str_coord

    @staticmethod
    def initializeGraph(roads, roi=None):
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')

        if roi is None:
            _roads = roads
            geometries = [geom for geom in roads.geometry]
        else:
            ids = roads.sindex.intersection(roi)
            geometries = [roads.loc[_id].geometry for _id in ids]
            _roads = GeoDataFrame([{'geometry': geom} for geom in geometries], crs=roads.crs)

        graph, ciVertices, icVertices, edges = AbstractUrbanGraphLib.buildGraph(geometries)
        return _roads, graph, ciVertices, icVertices, edges

    @staticmethod
    def addEndPoint(roads, spatialIdx, graph, ciVertices, icVertices, edges, endPoint):
        buffDist = 40.0
        minDist, nearestPoint, nearestRow = GeomLib.getNearestFeature(
            roads, spatialIdx, endPoint, buffDist)
        nearestLine = nearestRow.geometry

        if (0 < minDist):
            AbstractUrbanGraphLib.addEdge(graph, ciVertices, icVertices, edges, LineString([nearestPoint, endPoint]))

        # Cuts the nearestLine in two
        newEdges = GeomLib.cutsLineStringByCurvilinearDistance(
            nearestLine, nearestLine.project(nearestPoint))
        for _edge in newEdges:
            AbstractUrbanGraphLib.addEdge(graph, ciVertices, icVertices, edges, _edge)

        return ciVertices[AbstractUrbanGraphLib.hashCoord([endPoint.x, endPoint.y])], nearestLine

    @staticmethod
    def backupUrbanGraph(graph, ciVertices, icVertices, edges):
        _graph = copy.deepcopy(graph)
        _ciVertices = copy.deepcopy(ciVertices) if ciVertices is not None else None 
        _icVertices = copy.deepcopy(icVertices) if icVertices is not None else None
        _edges = copy.deepcopy(edges) if edges is not None else None
        return _graph, _ciVertices, _icVertices, _edges



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame

from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.UrbanGraphLibOld import UrbanGraphLibOld


class STToRoadsSections(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, withoutCulDeSac=False):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        self.withoutCulDeSac = withoutCulDeSac

    def run(self):
        ga = UrbanGraphLibOld()
        ga.add([row for _, row in self.inputGdf.iterrows()])

        if self.withoutCulDeSac:
            rs = ga.getUniqueRoadsSectionsWithoutCulDeSac()
        else:
            rs = ga.getUniqueRoadsSections()

        return GeoDataFrame(rs, crs=self.inputGdf.crs)



'''
Created on 21 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, pi, sin
from shapely.geometry import Point, Polygon
from shapely.ops import unary_union
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.ellipse.EllipseLib import EllipseLib


class ShadowLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __projectPointOntoShadowPlane(pA, radDir, altitudeOfShadowPlane):
        # M point one line (AB) iff: pM = (1-k).pA + k.pB
        pB = Point((pA.x + radDir[0], pA.y + radDir[1], pA.z + radDir[2]))
        if (pA.z == pB.z):
            if (altitudeOfShadowPlane == pA.z):
                return pA
            return None
        else:
            k = (altitudeOfShadowPlane - pA.z) / (pB.z - pA.z)
            xM = (1 - k) * pA.x + k * pB.x 
            yM = (1 - k) * pA.y + k * pB.y
            zM = altitudeOfShadowPlane
            return Point((xM, yM, zM))

    @staticmethod
    def __projectWallOntoShadowPlane(pA, pB, occluderElevation, radDir,
                                     altitudeOfShadowPlane):
        ppA = ShadowLib.__projectPointOntoShadowPlane(
            Point((pA.x, pA.y, occluderElevation)), radDir, altitudeOfShadowPlane)
        ppB = ShadowLib.__projectPointOntoShadowPlane(
            Point((pB.x, pB.y, occluderElevation)), radDir, altitudeOfShadowPlane)
        return Polygon([ (pA.x, pA.y), (pB.x, pB.y), (ppB.x, ppB.y), (ppA.x, ppA.y) ])

    @staticmethod
    def __projectRingOntoShadowPlane(ring, occluderElevation, exteriorRing,
                                     radDir, altitudeOfShadowPlane):
        if (0.0 < ring.length):
            n, shadows = len(list(ring.coords)), [Polygon(ring)] if exteriorRing else []
            for i in range(1, n):
                pA = Point(ring.coords[i - 1])
                pB = Point(ring.coords[i])
                shadows.append(ShadowLib.__projectWallOntoShadowPlane(
                    pA, pB, occluderElevation, radDir, altitudeOfShadowPlane))
            return unary_union(shadows)
        return None

    @staticmethod
    def projectBuildingOntoShadowPlane(polygon, occluderElevation, radDir, altitudeOfShadowPlane):
        _ring = polygon.exterior 
        shadow = ShadowLib.__projectRingOntoShadowPlane(
            _ring, occluderElevation, True, radDir, altitudeOfShadowPlane) 

        # Handle hole(s)
        for _ring in polygon.interiors:
            _shadow = ShadowLib.__projectRingOntoShadowPlane(
                _ring, occluderElevation, False, radDir, altitudeOfShadowPlane)
            if GeomLib.isHoled(_shadow):
                for _hole in _shadow.interiors:
                    shadow = shadow.difference(Polygon(_hole))

        return shadow

    @staticmethod
    def projectSphericalTreeOntoShadowPlane(treePosition, treeHeight, treeCrownRadius,
                                            treeTrunkRadius, radDir, solarAlti, solarAzim,
                                            altitudeOfShadowPlane, npoints):
        # Tree crown        
        # the projection of a sphere on a horizontal plane is an ellipse
        a = treeCrownRadius
        b = a / sin(solarAlti)
        c = Point((treePosition.x, treePosition.y, treeHeight - a))
        pc = Point(ShadowLib.__projectPointOntoShadowPlane(c, radDir, altitudeOfShadowPlane))
        treeCrownShadow = EllipseLib.getEllipseContour(pc, a, b, solarAzim, npoints)

        # Tree trunk
        if treeTrunkRadius is None:
            # Extrait de 
            # https://patternformation.wordpress.com/2013/01/08/la-hauteur-des-arbres-2/
            # h = 90.1 * a**(2/3)
            # a = (h / 90.1)**(3/2)
            a = max(0.1, (treeHeight / 90.1) ** (3 / 2))
        else:
            a = treeTrunkRadius
        c = Point((treePosition.x, treePosition.y, treeHeight - 2 * a))
        p1 = Point(c.x + a * cos(solarAzim + pi / 2), c.y + a * sin(solarAzim + pi / 2), c.z)
        p2 = Point(c.x + a * cos(solarAzim - pi / 2), c.y + a * sin(solarAzim - pi / 2), c.z)
        pp1 = ShadowLib.__projectPointOntoShadowPlane(p1, radDir, altitudeOfShadowPlane)
        pp2 = ShadowLib.__projectPointOntoShadowPlane(p2, radDir, altitudeOfShadowPlane)
        treeTrunkShadow = Polygon([(p1.x, p1.y, 0), (p2.x, p2.y, 0), pp2, pp1])

        return treeCrownShadow.union(treeTrunkShadow)

    @staticmethod
    def projectConicalTreeOntoShadowPlane(treePosition, treeHeight, treeCrownHeight,
                                          treeUpperCrownRadius, treeLowerCrownRadius,
                                          treeTrunkRadius, radDir, solarAlti, solarAzim,
                                          altitudeOfShadowPlane, npoints):
        # Tree crown
        cUp = Point((treePosition.x, treePosition.y, treeHeight))
        cDown = Point((treePosition.x, treePosition.y, treeHeight - treeCrownHeight))
        a, b = treeUpperCrownRadius, treeLowerCrownRadius

        incAngle, deltaAngle = 0, 2 * pi / npoints
        pPrevUp, pPrevDown = None, None

        treeCrownShadows, pUpperNodes = [], []
        for i in range(npoints + 1):
            _cos, _sin = cos(incAngle), sin(incAngle)
            currUp = Point((cUp.x + a * _cos, cUp.y + a * _sin, cUp.z))
            currDown = Point((cDown.x + b * _cos, cDown.y + b * _sin, cDown.z))

            pCurrUp = Point(ShadowLib.__projectPointOntoShadowPlane(
                currUp, radDir, altitudeOfShadowPlane))
            pCurrDown = Point(ShadowLib.__projectPointOntoShadowPlane(
                currDown, radDir, altitudeOfShadowPlane))

            pUpperNodes.append(pCurrUp)

            if 0 < i:
                treeCrownShadows.append(Polygon((pPrevUp, pCurrUp, pCurrDown, pPrevDown))) 

            pPrevUp, pPrevDown = pCurrUp, pCurrDown
            incAngle += deltaAngle
        treeCrownShadows.append(Polygon(pUpperNodes))
        treeCrownShadow = unary_union(treeCrownShadows)

        # Tree trunk
        if treeTrunkRadius is None:
            # Extrait de 
            # https://patternformation.wordpress.com/2013/01/08/la-hauteur-des-arbres-2/
            # h = 90.1 * a**(2/3)
            # a = (h / 90.1)**(3/2)
            a = max(0.1, (treeHeight / 90.1) ** (3 / 2))
        else:
            a = treeTrunkRadius
        c = Point((treePosition.x, treePosition.y, treeHeight - treeCrownHeight))
        p1 = Point(c.x + a * cos(solarAzim + pi / 2), c.y + a * sin(solarAzim + pi / 2), c.z)
        p2 = Point(c.x + a * cos(solarAzim - pi / 2), c.y + a * sin(solarAzim - pi / 2), c.z)
        pp1 = ShadowLib.__projectPointOntoShadowPlane(p1, radDir, altitudeOfShadowPlane)
        pp2 = ShadowLib.__projectPointOntoShadowPlane(p2, radDir, altitudeOfShadowPlane)
        treeTrunkShadow = Polygon([(p1.x, p1.y, 0), (p2.x, p2.y, 0), pp2, pp1])

        _result = treeCrownShadow.union(treeTrunkShadow)
        # Use a buffer to avoid slivers
        _result = _result.buffer(0.001, -1)

        return _result



'''
Created on 20 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STSpatialJoin(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdfLeft, spatialRelationship, inputGdfRight):
        '''
        Constructor
        '''
        if not isinstance(inputGdfLeft, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdfLeft, 'GeoDataFrame')
        if not isinstance(inputGdfRight, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdfRight, 'GeoDataFrame')
        if not (inputGdfLeft.crs == inputGdfRight.crs):
            raise Exception('Illegal argument: both GeoDataFrame must share shames CRS!')

        self.inputGdfLeft = inputGdfLeft
        self.inputGdfRight = inputGdfRight

        if ('contains' == spatialRelationship):
            self.query = lambda geoms: geoms[0].contains(geoms[1])
        elif ('crosses' == spatialRelationship):
            self.query = lambda geoms: geoms[0].crosses(geoms[1])
        elif 'intersects' == spatialRelationship:
            self.query = lambda geoms: geoms[0].intersects(geoms[1])
        elif 'overlaps' == spatialRelationship:
            self.query = lambda geoms: geoms[0].overlaps(geoms[1])
        elif 'touches' == spatialRelationship:
            self.query = lambda geoms: geoms[0].touches(geoms[1])
        elif ('within' == spatialRelationship):
            self.query = lambda geoms: geoms[0].within(geoms[1])
        elif ('contains_centroid' == spatialRelationship):
            self.query = lambda geoms: geoms[0].contains(geoms[1].centroid)
        else:
            raise Exception('Unknown spatial relationship! It has to be chosen among: \
            "contains", "crosses", "intersects", "overlaps", \
            "touches", "within", and "contains_centroid".')

    def run(self):
        spatialIdx = self.inputGdfRight.sindex

        rows = []
        for _, rowLeft in self.inputGdfLeft.iterrows():
            geomLeft = rowLeft.geometry
            bbLeft = geomLeft.bounds

            idsRight = spatialIdx.intersection(bbLeft)
            for idRight in idsRight:
                rowRight = self.inputGdfRight.loc[idRight]
                geomRight = rowRight.geometry
                if self.query([geomLeft, geomRight]):
                    rows.append(self.appendNewItems(rowLeft, rowRight))

        return GeoDataFrame(rows, crs=self.inputGdfLeft.crs)



'''
Created on 23 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import arctan, cos, pi, sin
from shapely.geometry import Polygon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class SkyMap2D(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, nRays=180, size=4.0, maxRayLen=100.0, elevationFieldname='HAUTEUR',
                 projectionName='Stereographic'):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf
        self.spatialIndex = buildingsGdf.sindex

        self.nRays = nRays
        self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.size = size
        self.maxRayLen = maxRayLen

        if elevationFieldname not in buildingsGdf:
            raise Exception('%s is not a relevant field name!' % elevationFieldname)
        self.elevationFieldname = elevationFieldname

        if not projectionName in ['Stereographic']:
            raise IllegalArgumentTypeException(projectionName, 'spherical projection as "Stereographic"')
        self.proj = SkyMap2D.__stereographic

    @staticmethod
    def __stereographic(lat, lon, size=1.0):
        radius = (size * cos(lat)) / (1.0 + sin(lat))
        return (radius * cos(lon), radius * sin(lon))

    def runWithArgs(self, row):
        viewPoint = row.geometry.centroid

        if GeomLib.isAnIndoorPoint(viewPoint, self.buildingsGdf, self.spatialIndex):
            return Polygon()

        _, _, _, _, hws = RayCastingLib.multipleRayCast25D(self.buildingsGdf, self.spatialIndex, viewPoint,
                                                           self.shootingDirs, self.maxRayLen, self.elevationFieldname, True)
        lats = [arctan(hw) for hw in hws]
        angularOffset = (2.0 * pi) / self.nRays
        lons = [i * angularOffset for i in list(range(self.nRays)) + [0]]

        pnodes = [self.proj(lats[i], lons[i], self.size) for i in list(range(self.nRays)) + [0]]
        nodes = [(viewPoint.x + pnode[0], viewPoint.y + pnode[1]) for pnode in pnodes]

        return { 'geometry': Polygon(viewPoint.buffer(self.size).exterior.coords, [nodes]) }



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import array, log10, sqrt, sum


class AbstractDistance(object):
    '''
    classdocs
    '''

    def __init__(self, distName='Euclidean'):
        if ('Canberra' == distName):
            self.distTool = CanberraDistance()
        elif ('Chebyshev' == distName):
            self.distTool = ChebyshevDistance()
        elif ('EarthMovers' == distName):
            self.distTool = EarthMoversDistance()
        elif ('Euclidean' == distName):
            self.distTool = EuclideanDistance()
        elif ('Manhattan' == distName):
            self.distTool = ManhattanDistance()
        elif ('DTW' == distName):
            self.distTool = DTWDistance()
        elif ('SNR' == distName):
            self.distTool = SNRDistance()
        else:
            raise Exception('Unknown distance. Distance, if mentioned, must be ' + 
                            'chosen among: Canberra, Chebyshev, EarthMovers, ' + 
                            'Euclidean (default), Manhattan, DTW, and SNR!')
        '''
        elif ('Hausdorff' == distName):
            self.distTool = HausdorffDistance()
        '''

    def compute(self, lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
            # raise Exception('The two lists must be of the same size!')
        return self.distTool.compute(lst1, lst2)


class CanberraDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
        acc = 0.0
        for i in range(len(lst1)):
            num = abs(lst1[i] - lst2[i])
            denom = abs(lst1[i]) + abs(lst2[i])
            tmp = 0 if ((0 == num) and (0 == denom)) else num / denom
            acc += tmp
        return acc


class ChebyshevDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
        maxim = 0.0
        for i in range(len(lst1)):
            maxim = max(maxim, abs(lst1[i] - lst2[i]))
        return maxim


class EarthMoversDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
        lastDistance = 0.0
        totalDistance = 0.0
        for i in range(len(lst1)):
            currentDistance = (lst1[i] + lastDistance) - lst2[i]
            totalDistance += abs(currentDistance)
            lastDistance = currentDistance
        return totalDistance


class EuclideanDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
        acc = 0.0
        for i in range(len(lst1)):
            tmp = lst1[i] - lst2[i]
            acc += tmp * tmp
        return sqrt(acc)


class ManhattanDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
        acc = 0.0
        for i in range(len(lst1)):
            acc += abs(lst1[i] - lst2[i])
        return acc

'''
class HausdorffDistance(object):
    @staticmethod
    def compute(lst1, lst2):
        if (len(lst1) != len(lst2)):
            return float('inf')
        cartc1 = PolarCartesianCoordinates.fromRayLengthsToCartesianCoordinates(lst1)
        cartc2 = PolarCartesianCoordinates.fromRayLengthsToCartesianCoordinates(lst2)
        dists = []
        for c1 in cartc1:
            dists.append(min([EuclideanDistance.compute(c1, c2) for c2 in cartc2]))
        for c2 in cartc2:
            dists.append(min([EuclideanDistance.compute(c1, c2) for c1 in cartc1]))
        return max(dists)
        # raise NotImplementedError('HausdorffDistance must be implemented first!')
'''


class DTWDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        nc = len(lst1)
        nl = len(lst2)
        dtw = [[ float('inf') for _ in range(nc + 1) ] for _ in range(nl + 1) ]
        dtw[0][0] = 0.0
        for l in range(nl):
            for c in range(nc):
                cost = abs(lst1[c] - lst2[l])
                dtw[l + 1][c + 1] = cost + min(dtw[l][c + 1], dtw[l + 1][c], dtw[l][c])
        return dtw[nl][nc]


class SNRDistance(object):

    @staticmethod
    def compute(lst1, lst2):
        '''
        To compare the performance of the recovery algorithms, the SNR (Signal to noise ratio) will be used.
        Original signal: lst1
        Restored signal: lst2
        '''
        if (len(lst1) != len(lst2)):
            return float('inf')
        t1, t2 = array(lst1), array(lst2)
        return 10 * log10(sum(t1 ** 2) / sum((t1 - t2) ** 2))

# for d in ['Canberra', 'Chebyshev', 'EarthMovers', 'Euclidean', 'Manhattan', 'DTW']:
#     print AbstractDistance(d).compute([1, 1, 1, 1], [1, 1, 1, 2])

# print CanberraDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])
# print ChebyshevDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])
# print EarthMoversDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])
# print EuclideanDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])
# print ManhattanDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])
# print HausdorffDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])
# print DTWDistance.compute([1, 1, 1, 1], [1, 1, 1, 2])



'''
Created on 10 sept. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeoProcess import GeoProcess

from qgis.core import QgsLayerTreeLayer, QgsProject


class ShowFeatureCount(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, layerNames):
        '''
        Constructor
        '''
        self.layerNames = layerNames

    def run(self):
        root = QgsProject.instance().layerTreeRoot()
        for child in root.children():
            if isinstance(child, QgsLayerTreeLayer):
                if child.name() in self.layerNames:
                    child.setCustomProperty("showFeatureCount", True)



'''
Created on 22 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas.core.common import flatten


class ListUtilities(object):
    '''
    classdocs
    '''

    @staticmethod
    def isASubList(smallList, bigList):
        return all(i in bigList for i in smallList)
        
    @staticmethod
    def isASubSeq(smallList, bigList):
        if (len(bigList) < len(smallList)):
            return False
        for i in range(len(bigList)):            
            if smallList == ListUtilities.rotate(bigList, i)[:len(smallList)]:
                return True
        return False

    @staticmethod
    def flatten(myList):
        return list(flatten(myList))

    @staticmethod
    def lengthOfTheLongestConsecutiveSeq(seq):
        seq = sorted(seq)
        prev, currLen, longestLen = None, 0, 0
        for curr in seq:
            currLen += 1
            if prev is not None:
                if 1 < (curr - prev):
                    longestLen = max(longestLen, currLen - 1)
                    currLen = 1
            prev = curr
        return max(longestLen, currLen)

    @staticmethod
    def rotate(seq, n):
        n = n % len(seq)
        return seq[n:] + seq[:n]

    @staticmethod
    def allRotations(seq):
        if seq is None:
            return None
        return [ListUtilities.rotate(seq, n) for n in range(len(seq))]



'''
Created on 21 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone, timedelta
from warnings import warn

from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.PySolarSunLib import PySolarSunLib
from t4gpd.commons.sun.SoleneSunLib import SoleneSunLib

import matplotlib.pyplot as plt


class SunLib(object):
    '''
    classdocs
    '''

    def __init__(self, gdf=LatLonLib.NANTES, model='pysolar'):
        '''
        Constructor
        '''
        if ('pysolar' == model.lower()):
            self.model = PySolarSunLib(gdf)
        elif ('solene' == model.lower()):
            warn('Select SOLENE model')
            self.model = SoleneSunLib(gdf)
        else:
            raise IllegalArgumentTypeException(model, '{"pysolar", "solene"}')

        self.lat, self.lon = LatLonLib.fromGeoDataFrameToLatLon(gdf)

    def getDayLengthInMinutes(self, dt):
        return self.model.getDayLengthInMinutes(dt)

    def getFractionalYear(self, dt):
        return self.model.getFractionalYear(dt)

    def getRadiationDirection(self, dt):
        return self.model.getRadiationDirection(dt)

    def getSolarAnglesInDegrees(self, dt):
        return self.model.getSolarAnglesInDegrees(dt)

    def getSolarAnglesInRadians(self, dt):
        return self.model.getSolarAnglesInRadians(dt)

    def getSolarDeclination(self, dayOfYear):
        return self.model.getSolarDeclination(dayOfYear)

    def getSunrise(self, dt):
        return self.model.getSunrise(dt)

    def getSunset(self, dt):
        return self.model.getSunset(dt)

    def plotDayLengths(self):
        year = datetime.now().year
        daysInYear = range(1, self.model.nDaysPerYear(year) + 1, 10)
        dayLengths = []

        for _dayOfYear in daysInYear:
            _dt = datetime(year, 1, 1, 12, tzinfo=timezone.utc) + timedelta(days=_dayOfYear)
            _dayLengthInMin = self.model.getDayLengthInMinutes(_dt)
            dayLengths.append(_dayLengthInMin)

        if (isinstance(self.model, SoleneSunLib)):
            _title = 'Solene model'
        elif (isinstance(self.model, PySolarSunLib)):
            _title = 'PySolar implementation'

        plt.title('%s: days lengths - Year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$' % (
            _title, year, self.lat, self.lon))
        plt.xlabel('Day of year')
        plt.ylabel('Day length (min.)')
        plt.grid()
        plt.plot(daysInYear, dayLengths)
        plt.show()      

    def plotSolarDeclination(self):
        _year = datetime.now().year
        _the1stOfJan = datetime(_year, 1, 1, 12, tzinfo=timezone.utc)

        days = range(1, self.model.nDaysPerYear(_year) + 1)

        declinations = [
            self.model.getSolarDeclination(_the1stOfJan + timedelta(days=dayOfYear)) 
            for dayOfYear in days]

        if (isinstance(self.model, SoleneSunLib)):
            plt.title('Solar declination by Solene (Spencer, 1971) - year = %d' % _year)
        elif (isinstance(self.model, PySolarSunLib)):
            plt.title('Solar declination by PySolar - year = %d' % _year)

        plt.xlabel('Day of year')
        plt.ylabel('Solar declination')
        plt.grid()
        plt.plot(days, declinations)
        plt.show()      

    def plotSunAltiAtNoon(self):
        _year = datetime.now().year
        _the1stOfJan = datetime(_year, 1, 1, 12, tzinfo=timezone.utc)

        daysInYear = range(1, self.model.nDaysPerYear(_year) + 1, 10)
        sunAltiAtNoon = []

        for _dayOfYear in daysInYear:
            _dt = _the1stOfJan + timedelta(days=_dayOfYear)
            _alti, _ = self.model.getSolarAnglesInDegrees(_dt)
            sunAltiAtNoon.append(_alti)

        if (isinstance(self.model, SoleneSunLib)):
            _title = 'Solene model'
        elif (isinstance(self.model, PySolarSunLib)):
            _title = 'PySolar implementation'

        plt.title('%s: days lengths - Year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$' % (
            _title, _year, self.lat, self.lon))
        plt.xlabel('Day of year')
        plt.ylabel('Solar altitude at noon (in degrees)')
        plt.grid()
        plt.plot(daysInYear, sunAltiAtNoon)
        plt.show()      

    def plotSunriseSunset(self):
        year = datetime.now().year
        daysInYear = range(1, self.model.nDaysPerYear(year) + 1, 10)
        sunrises, sunsets = [], []

        for _dayOfYear in daysInYear:
            _dt = datetime(year, 1, 1, 12, tzinfo=timezone.utc) + timedelta(days=_dayOfYear)

            _sunrise = self.model.getTimeSpentSinceMidnight(
                self.model.getSunrise(_dt))
            _sunset = self.model.getTimeSpentSinceMidnight(
                self.model.getSunset(_dt))

            sunrises.append(_sunrise)
            sunsets.append(_sunset)

        if (isinstance(self.model, SoleneSunLib)):
            _title = 'Solene model'
        elif (isinstance(self.model, PySolarSunLib)):
            _title = 'PySolar implementation'

        plt.title('%s: sunrise/sunset - Year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$' % (
            _title, year, self.lat, self.lon))
        plt.xlabel('Day of year')
        plt.ylabel('UTC Time (hr.)')
        plt.grid()
        plt.plot(daysInYear, sunrises, daysInYear, sunsets)
        plt.show()      

    def plotSolarPanorama(self):
        year = datetime.now().year
        _, basemap = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))

        if (isinstance(self.model, SoleneSunLib)):
            _title = 'Solene model'
        elif (isinstance(self.model, PySolarSunLib)):
            _title = 'PySolar implementation'
        plt.title('%s: solar panorama - year %d\nLatitude: %.1f$^o$, Longitude: %.1f$^o$' % (
            _title, year, self.lat, self.lon))
        plt.xlabel('Azim. [$^0$]')
        plt.ylabel('Alti. [$^0$]')

        altis, azims = {}, {}
        hrsAltis, hrsAzims = {}, {}

        for hour in range(0, 24):
            hrsAltis[hour], hrsAzims[hour] = [], []

        for month in range(6, 13, 3):
            altis[month], azims[month] = [], [] 
            t0 = datetime(year, month, 21, tzinfo=timezone.utc)
            _sunrise = self.model.getSunrise(t0)
            _sunset = self.model.getSunset(t0)

            for hour in range(0, 24):
                for minute in range(0, 60, 10):
                    _dt = t0 + timedelta(hours=hour) + timedelta(minutes=minute)
                    if True or (_sunrise < _dt < _sunset):
                        _alti, _azim = self.model.getNativeSolarAnglesInDegrees(_dt)
                        if (0 <= _alti):
                            altis[month].append(_alti)
                            azims[month].append(_azim)
                        if (0 == minute):
                            hrsAltis[hour].append(_alti)
                            hrsAzims[hour].append(_azim)

            basemap.plot(azims[month], altis[month], label=_dt.strftime('%b %d'))

        for hour in range(5, 20, 1):
            if (0 == (hour % 2)):
                basemap.plot(hrsAzims[hour], hrsAltis[hour], 'k-.')
                basemap.text((3 * hrsAzims[hour][0] + hrsAzims[hour][-1]) / 4,
                              (3 * hrsAltis[hour][0] + hrsAltis[hour][-1]) / 4,
                              f'{hour}:00', fontsize=9).set_color('black')

        for lbl, azim in [('East', 90), ('South', 180), ('West', 270)]:
            basemap.axvline(x=azim, linestyle=':', color='black', linewidth=1)
            basemap.text(azim, 70, lbl, fontsize=9).set_color('black')
        basemap.set_ylim(0, 90)

        basemap.legend()
        plt.show()      



'''
Created on 21 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timedelta

from numpy import cos, round, sin
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.LatLonLib import LatLonLib


class AbstractSunLib(object):
    '''
    classdocs
    '''

    def __init__(self, gdf=LatLonLib.NANTES):
        self.lat, self.lon = LatLonLib.fromGeoDataFrameToLatLon(gdf)

    def __fromSolarAnglesToRadiationDirection(self, alti, azim):
        cosAlt = cos(alti)
        x = cosAlt * cos(azim)
        y = cosAlt * sin(azim)
        z = sin(alti)
        return x, y, z

    def getDayLengthInMinutes(self, dt):
        _sunrise = self.getSunrise(dt)
        _sunset = self.getSunset(dt)
        return (_sunset - _sunrise).seconds // 60

    def getDayOfYear(self, dt):
        return dt.timetuple().tm_yday

    def getMinutesSpentSinceMidnight(self, dt):
        _midnight = dt.replace(hour=0, minute=0, second=0, microsecond=0)
        return int(round((dt - _midnight) / timedelta(minutes=1)))

    def getTimeSpentSinceMidnight(self, dt):
        _midnight = dt.replace(hour=0, minute=0, second=0, microsecond=0)
        return ((dt - _midnight) / timedelta(hours=1))

    def getRadiationDirection(self, dt):
        alti, azim = self.getSolarAnglesInRadians(dt)
        return self.__fromSolarAnglesToRadiationDirection(alti, azim)

    def getSolarAnglesInDegrees(self, dt):
        raise NotImplementedError('getSolarAnglesInDegrees(...) must be overridden!')

    def getSolarAnglesInRadians(self, dt):
        alti, azim = self.getSolarAnglesInDegrees(dt)
        return AngleLib.toRadians(alti), AngleLib.toRadians(azim)

    def getSolarDeclination(self, dayOfYear):
        raise NotImplementedError('getSolarDeclination(...) must be overridden!')

    def isALeapYear(self, year):
        if isinstance(year, datetime):
            year = year.year
        return (0 == year % 400) if (0 == year % 100) else (0 == year % 4)

    def nDaysPerYear(self, year):
        if isinstance(year, datetime):
            year = year.year
        return 366 if (self.isALeapYear(year)) else 365



'''
Created on 6 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LineString
from shapely.ops import substring

from geopandas.geodataframe import GeoDataFrame
from numpy import abs, round
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STGradientOfDistancesToBuildings(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, lines, buildings, sampleDist, pathidFieldname=None, threshold=0.1, order=1):
        '''
        Constructor
        '''
        if not isinstance(lines, GeoDataFrame):
            raise IllegalArgumentTypeException(lines, 'GeoDataFrame')
        self.lines = lines

        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings
        self.spatialIdx = buildings.sindex

        if (pathidFieldname is not None) and (pathidFieldname not in lines):
            raise Exception('%s is not a relevant field name!' % (pathidFieldname))

        self.sampleDist = sampleDist
        self.pathidFieldname = pathidFieldname
        self.threshold = threshold

        if not order in (1, 2, 3):
            raise IllegalArgumentTypeException(order, '(1, 2, 3)')
        self.fieldToTest = 'r_deriv%d' % (order)

    def __getType(self, _nodeRow):
        _value = _nodeRow[self.fieldToTest]
        if (_value is not None) and (self.threshold > abs(_value)):
            return 'canyon'
        return 'square'
        
    def run(self):

        nodesRows, segmentsRows, linesRows = [], [], []
        for _id, row in self.lines.iterrows():
            _line = row.geometry
            _lineLen = _line.length
            if (self.sampleDist < _lineLen):
                _pathid = _id if (self.pathidFieldname is None) else row[self.pathidFieldname] 
                _nSegm = int(round(_lineLen / self.sampleDist))
                _sampleDist = _lineLen / _nSegm

                _nodesRows, _segmentsRows = [], []
                for i in range(0, _nSegm + 1):
                    _sampleGeom = _line.interpolate(i * _sampleDist)
                    _r, _nearestPoint, _ = GeomLib.getNearestFeature(
                        self.buildings, self.spatialIdx, _sampleGeom, buffDist=40.0)
                    _nodesRows.append({
                        'pathid': _pathid,
                        'nodeid': i,
                        'geometry': _sampleGeom,
                        'curv_absc': i * _sampleDist,
                        'r': _r,
                        'r_deriv1': None,
                        'r_deriv2': None,
                        'r_deriv3': None,
                        'area_type': None
                        })

                    _segmentsRows.append({
                        'pathid': _pathid,
                        'nodeid': i,
                        'geometry': LineString([(_sampleGeom.x, _sampleGeom.y), (_nearestPoint.x, _nearestPoint.y)]),
                        'curv_absc': i * _sampleDist,
                        'r': _r,
                        'r_deriv1': None,
                        'r_deriv2': None,
                        'r_deriv3': None,
                        'area_type': None
                        })

                # Assess 1st and 2nd order derivatives
                for i in range(1, _nSegm):
                    _hp1 = _nodesRows[i + 1]['geometry'].distance(_nodesRows[i]['geometry'])
                    _hm1 = _nodesRows[i]['geometry'].distance(_nodesRows[i - 1]['geometry'])

                    up1, u0, um1 = _nodesRows[i + 1]['r'], _nodesRows[i]['r'], _nodesRows[i - 1]['r']
                    
                    _fwdDiff = (up1 - u0) / _hp1
                    _bwdDiff = (u0 - um1) / _hm1
                    _nodesRows[i]['r_deriv1'] = (_fwdDiff + _bwdDiff) / 2.0
                    _cenDiff = (up1 - 2 * u0 + um1) / (_hp1 * _hm1)
                    _nodesRows[i]['r_deriv2'] = _cenDiff

                    if (1 < i < _nSegm - 1):
                        up2, um2 = _nodesRows[i + 2]['r'], _nodesRows[i - 2]['r']
                        _hp2 = _nodesRows[i + 2]['geometry'].distance(_nodesRows[i + 1]['geometry'])
                        _hm2 = _nodesRows[i - 2]['geometry'].distance(_nodesRows[i - 1]['geometry'])

                        _fwdDiff2 = (up2 - 3 * up1 + 3 * u0 - um1) / (_hp2 * _hp1 * _hm1)
                        _bwdDiff2 = (up1 - 3 * u0 + 3 * um1 - um2) / (_hp1 * _hm1 * _hm2)

                        _nodesRows[i]['r_deriv3'] = (_fwdDiff2 + _bwdDiff2) / 2.0
                        _segmentsRows[i]['r_deriv3'] = _nodesRows[i]['r_deriv3']

                    elif (1 < i):
                        um2 = _nodesRows[i - 2]['r']
                        _hm2 = _nodesRows[i - 2]['geometry'].distance(_nodesRows[i - 1]['geometry'])

                        _bwdDiff2 = (up1 - 3 * u0 + 3 * um1 - um2) / (_hp1 * _hm1 * _hm2)

                        _nodesRows[i]['r_deriv3'] = _bwdDiff2
                        _segmentsRows[i]['r_deriv3'] = _nodesRows[i]['r_deriv3']

                    elif (i < _nSegm - 1):
                        up2 = _nodesRows[i + 2]['r']
                        _hp2 = _nodesRows[i + 2]['geometry'].distance(_nodesRows[i + 1]['geometry'])

                        _fwdDiff2 = (up2 - 3 * up1 + 3 * u0 - um1) / (_hp2 * _hp1 * _hm1)

                        _nodesRows[i]['r_deriv3'] = _fwdDiff2
                        _segmentsRows[i]['r_deriv3'] = _nodesRows[i]['r_deriv3']

                    _nodesRows[i]['area_type'] = self.__getType(_nodesRows[i])

                    _segmentsRows[i]['r_deriv1'] = _nodesRows[i]['r_deriv1']
                    _segmentsRows[i]['r_deriv2'] = _nodesRows[i]['r_deriv2']
                    _segmentsRows[i]['area_type'] = _nodesRows[i]['area_type']

                _nodesRows[0]['r_deriv1'] = (_nodesRows[1]['r'] - _nodesRows[0]['r']) / _nodesRows[1]['geometry'].distance(_nodesRows[0]['geometry'])
                _nodesRows[0]['area_type'] = self.__getType(_nodesRows[0])

                _segmentsRows[0]['r_deriv1'] = _nodesRows[0]['r_deriv1']
                _segmentsRows[0]['area_type'] = _nodesRows[0]['area_type']

                _nodesRows[-1]['r_deriv1'] = (_nodesRows[-1]['r'] - _nodesRows[-2]['r']) / _nodesRows[-1]['geometry'].distance(_nodesRows[-2]['geometry'])
                _nodesRows[-1]['area_type'] = self.__getType(_nodesRows[-1])

                _segmentsRows[-1]['r_deriv1'] = _nodesRows[-1]['r_deriv1']
                _segmentsRows[-1]['area_type'] = _nodesRows[-1]['area_type']

                nodesRows += _nodesRows
                segmentsRows += _segmentsRows

                # Extract "continuous" sub-linestrings
                _prevRow, _prevStatus = _nodesRows[0], self.__getType(_nodesRows[0])
                for i in range(1, _nSegm + 1):
                    _currRow = _nodesRows[i]
                    _currStatus = self.__getType(_currRow)

                    if (_prevStatus != _currStatus) or (_nSegm == i):
                        _subls = substring(_line, _prevRow['curv_absc'], _currRow['curv_absc'])
                        linesRows.append({
                            'pathid': _pathid,
                            'geometry': _subls,
                            'area_type': _prevStatus
                            })
                        _prevRow, _prevStatus = _currRow, _currStatus

        return [
            GeoDataFrame(nodesRows, crs=self.lines.crs),
            GeoDataFrame(segmentsRows, crs=self.lines.crs),
            GeoDataFrame(linesRows, crs=self.lines.crs)]



'''
Created on 17 sept. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import ndarray
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

from qgis.core import QgsRectangle, QgsVectorLayer
from qgis.utils import iface


class ZoomLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def setExtent(roi):
        if not isinstance(roi, QgsRectangle):
            if isinstance(roi, QgsVectorLayer):
                roi = roi.extent()
            elif isinstance(roi, GeoDataFrame):
                roi = QgsRectangle(*roi.total_bounds)
            elif GeomLib.isAShapelyGeometry(roi):
                roi = QgsRectangle(*roi.bounds)
            elif isinstance(roi, ndarray) and (4 == len(roi)):
                roi = QgsRectangle(*roi)
            else:
                raise IllegalArgumentTypeException(roi, 'QgsRectangle, QgsVectorLayer, GeoDataFrame or ndarray')

        iface.mapCanvas().setExtent(roi)
        iface.mapCanvas().refresh()
        return roi



'''
Created on 10 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry.linestring import LineString
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class LineStringCuttingLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __fromLineStringToSegments(line, cuttingDist):
        if not isinstance(line, LineString):
            raise IllegalArgumentTypeException(line, 'LineString')
        if not (2 == len(line.coords)):
            raise IllegalArgumentTypeException(line, '2-point LineString')

        result = []
        coords, lineLen = line.coords, line.length
        if (lineLen <= cuttingDist):
            result.append(line)
        else:
            _d = lineLen / round(lineLen / cuttingDist)
            _prev, _accD = coords[0], _d
            while (_accD < lineLen):
                _curr = line.interpolate(_accD, normalized=False)
                result.append(LineString([_prev, _curr]))
                _prev, _accD = _curr, _accD + _d
            result.append(LineString([_prev, coords[1]]))
        return result

    @staticmethod
    def cuttingIntoSegments(gdf, cuttingDist):
        result = []
        for _, row in gdf.iterrows():
            _geom = row.geometry
            _segms = GeomLib.toListOfBipointsAsLineStrings(_geom)
            for _segm in _segms:
                result += LineStringCuttingLib.__fromLineStringToSegments(_segm, cuttingDist)
        return GeoDataFrame([{'gid': i, 'geometry': g} for i, g in enumerate(result)], crs=gdf.crs)



'''
Created on 10 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import MultiPolygon, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.GeomLib3D import GeomLib3D
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class CirWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile, translate=False):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        self.outputFile = outputFile

        minx, miny, _, _ = self.inputGdf.total_bounds
        self.llc = [minx, miny] if translate else [0, 0]

    def __printCoordinates(self, coord, llc=[0, 0]):
        return '\t%g\t%g\t%g' % (coord[0] - llc[0], coord[1] - llc[1], coord[2])

    def __printHeader(self):
        return '%d\t%d\r\n%s' % (
            len(self.inputGdf),
            len(self.inputGdf),
            '\t99999\t99999\r\n' * 5)

    def __printGeometry(self, faceIdx, geometry):
        if GeomLib.isPolygonal(geometry):
            if isinstance(geometry, Polygon):
                nCtrs = 1
                _polygons = [geometry]

            elif isinstance(geometry, MultiPolygon):
                nCtrs = len(geometry.geoms)
                _polygons = geometry.geoms

            _normal = GeomLib3D.getFaceNormalVector(_polygons[0])
            result = 'f%d %d\r\n%s\r\n' % (faceIdx, nCtrs, self.__printCoordinates(_normal))
            for _geom in _polygons:
                result += self.__printPolygon(_geom)
            return result

        raise IllegalArgumentTypeException(geometry, 'Polygon or MultiPolygon')

    def __printPolygon(self, polygon):
        result = 'c%d\r\n' % (len(polygon.interiors))
        result += self.__printRing(polygon.exterior)
        for hole in polygon.interiors:
            result += 't\r\n'
            result += self.__printRing(hole)
        return result

    def __printRing(self, ring):
        ring = GeomLib.addZCoordinateIfNecessary(ring)
        return '%d\r\n%s\r\n' % (
            len(ring.coords),
            '\r\n'.join([self.__printCoordinates(c, self.llc) for c in ring.coords]))

    def run(self):
        with open(self.outputFile, 'w') as f:
            f.write(self.__printHeader())

            faceIdx = 1
            for _, row in self.inputGdf.iterrows():
                _geom = row.geometry
                f.write(self.__printGeometry(faceIdx, _geom))
                faceIdx += 1



'''
Created on 27 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString
from t4gpd.commons.BoundingBox import BoundingBox
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.UrbanGraphLibOld import UrbanGraphLibOld
from t4gpd.morph.STVoronoiPartition import STVoronoiPartition

from t4gpd.morph.STPointsDensifier import STPointsDensifier


class STSkeletonizeTheVoid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, samplingDist=10.0):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf
        self.spatialIdx = buildingsGdf.sindex

        self.samplingDist = samplingDist

    def run(self):
        roiGeom = BoundingBox(self.buildingsGdf).asPolygon()

        # Sample the building contours
        _nodesGdf = STPointsDensifier(self.buildingsGdf, self.samplingDist).run()

        # Voronoi Diagram
        _voronoiGdf = STVoronoiPartition(_nodesGdf).run()

        # Remove useless edges
        _ug = UrbanGraphLibOld()
        for _, row in _voronoiGdf.iterrows():
            _edges = GeomLib.toListOfLineStrings(row.geometry)
            for _edge in _edges:
                _prev = None
                for _curr in _edge.coords:
                    if _prev is not None:
                        keep = True
                        _currGeom = LineString((_prev, _curr))
                        _buildingsIds = self.spatialIdx.intersection(_currGeom.bounds)
                        for _buildingId in _buildingsIds:
                            _buildingGeom = self.buildingsGdf.loc[_buildingId].geometry
                            if _currGeom.intersects(_buildingGeom):
                                keep = False
                                break
                        if keep:
                            _ug.add(_currGeom.intersection(roiGeom))
                    _prev = _curr

        rows = _ug.getUniqueRoadsSections()
        return GeoDataFrame(rows, crs=self.buildingsGdf.crs)



'''
Created on 26 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone
from locale import LC_ALL, setlocale
from numpy import dot, sqrt
from numpy.linalg import norm
from t4gpd.commons.Epsilon import Epsilon
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.LatLonLib import LatLonLib
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.energy.PerrinDeBrichambaut import PerrinDeBrichambaut

import matplotlib.pyplot as plt


class DirectSolarIrradianceLib(object):
    '''
    classdocs

    Irradiance is understood as instantaneous density of solar radiation incident on a given
    surface, typically expressed in W/m2.

    Irradiation is the sum of irradiance over a time period (e.g. 1 hour, day, month, etc.)
    expressed in J/m2. However, in daily routine Wh/m2 are more commonly used.
    '''

    @staticmethod
    def noMaskDI(normal, dt, gdf=LatLonLib.NANTES, model='pysolar', skyType=PerrinDeBrichambaut.STANDARD_SKY):
        '''
        Direct Irradiance (DI in [W/m2]) is the amount of instantaneous solar radiation 
        received per unit area by a surface whose normal is given as an input parameter. 
        '''
        if not isinstance(normal, (list, tuple)) or (len(normal) not in [2, 3]):
            raise IllegalArgumentTypeException(normal, 'list or tuple of X, Y, and Z components')
        if not Epsilon.equals(1.0, norm(normal), 1e-3):
            raise IllegalArgumentTypeException(normal, 'unit vector')

        if not isinstance(dt, datetime):
            raise IllegalArgumentTypeException(dt, 'datetime')
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)

        normal = (*normal, 0.0) if (2 == len(normal)) else normal
        sunModel = SunLib(gdf=gdf, model=model)
        solarAlti, _ = sunModel.getSolarAnglesInRadians(dt)
        radDir = sunModel.getRadiationDirection(dt)
        
        dotProd = dot(radDir, normal)

        if (0.0 < solarAlti) and (0.0 < dotProd):
            dni = PerrinDeBrichambaut.directNormalIrradiance(solarAlti, skyType=skyType)
            di = dotProd * dni
            return di
        return 0.0

    @staticmethod
    def noMaskDNI(dt, gdf=LatLonLib.NANTES, model='pysolar', skyType=PerrinDeBrichambaut.STANDARD_SKY):
        '''
        Direct Normal Irradiance (DNI in [W/m2]) is the amount of solar radiation received 
        per unit area by a surface that is always held perpendicular (or normal) to the rays
        that come in a straight line from the direction of the sun at its current position in 
        the sky. 
        '''
        if not isinstance(dt, datetime):
            raise IllegalArgumentTypeException(dt, 'datetime')
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)

        sunModel = SunLib(gdf=gdf, model=model)
        solarAlti, _ = sunModel.getSolarAnglesInRadians(dt)

        if (0.0 < solarAlti):
            return PerrinDeBrichambaut.directNormalIrradiance(solarAlti, skyType=skyType)
        return 0.0

    @staticmethod
    def plot(label, normal, ofile=None):
        magn = 0.5

        lat, _ = LatLonLib.fromGeoDataFrameToLatLon(LatLonLib.NANTES)

        fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(3 * magn * 8.26, 1.3 * magn * 8.26))
        fig.suptitle(f'Direct Solar Irradiance, lat={lat:+.1f}°', size=20)
        for i, month in enumerate([3, 6, 12]):
            X, Y, Z = [], [], []
            for hour in range(24):
                for minutes in range(0, 60, 30):
                    dt = datetime(2021, month, 21, hour, minutes)
                    X.append(hour + minutes / 60)
                    Y.append(DirectSolarIrradianceLib.noMaskDI(normal, dt))
                    Z.append(DirectSolarIrradianceLib.noMaskDNI(dt))
            setlocale(LC_ALL, 'en_US.utf8')
            ax[i].set_title(f'{label} - {dt.strftime("%b %d")}st')
            ax[i].plot(X, Y, linestyle='-', color='black', linewidth=1)
            ax[i].plot(X, Z, linestyle=':', color='grey', linewidth=2, label='DNI')
            ax[i].set_xlabel('Hour')
            if (0 == i):
                ax[i].set_ylabel('Instantaneous irradiance [W.m$^{-2}$]')
            ax[i].set_ylim([0, 1000])
            ax[i].legend()
        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, format='pdf', bbox_inches='tight')
        plt.close(fig)

    @staticmethod
    def plotFelixMarboutin(ofile=None):
        t = 1.0 / sqrt(2.0)
        PAIRS = [
            # ('Roof', (0, 0, 1)),
            ('NE', (t, t, 0)),
            ('N', (0, 1, 0)),
            ('NW', (-t, t, 0)),
            ('W', (-1, 0, 0)),
            ('SW', (-t, -t, 0)),
            ('S', (0, -1, 0)),
            ('SE', (t, -t, 0)),
            ('E', (1, 0, 0)),
            ] 
        magn = 0.5

        lat, _ = LatLonLib.fromGeoDataFrameToLatLon(LatLonLib.NANTES)

        fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(3.0 * magn * 8.26, 1.1 * magn * 8.26))
        fig.suptitle(f'''Direct Solar Irradiance, lat={lat:+.1f}°''', size=20)
        MINS = range(0, 60, 15)
        for i, month in enumerate([12, 3, 6]):
            for label, normal in PAIRS:
                X, Y = [], []
                for hour in range(4, 22):
                    for minutes in MINS:
                        dt = datetime(2021, month, 21, hour, minutes)
                        X.append(hour + minutes / 60)
                        Y.append(DirectSolarIrradianceLib.noMaskDI(normal, dt))
                linestyle = '-' if (1 == len(label)) else ':'
                ax[i].plot(X, Y, linestyle=linestyle, linewidth=1.5,
                           label=f'{label} {sum(Y) / (1e3 * len(MINS)):.1f} kWh/m$^2$')

            setlocale(LC_ALL, 'en_US.utf8')
            ax[i].set_title(f'{dt.strftime("%b %d")}st')
            ax[i].set_xlabel('Hour')
            if (0 == i):
                ax[i].set_ylabel('Instantaneous irradiance [W.m$^{-2}$]')
                ax[i].set_xlim([8, 16])
            elif (1 == i):
                ax[i].set_xlim([6, 18])
            elif (1 == 2):
                ax[i].set_xlim([4, 20])
            ax[i].set_ylim([0, 900])
            ax[i].legend(loc='upper left', framealpha=0.5, ncol=2, fontsize=8)

        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, format='pdf', bbox_inches='tight')
        plt.close(fig)

    @staticmethod
    def plotFelixMarboutinAnnuel(ofile=None):
        t = 1.0 / sqrt(2.0)
        PAIRS = [
            # ('Roof', (0, 0, 1)),
            ('NE', (t, t, 0)),
            ('N', (0, 1, 0)),
            ('NW', (-t, t, 0)),
            ('W', (-1, 0, 0)),
            ('SW', (-t, -t, 0)),
            ('S', (0, -1, 0)),
            ('SE', (t, -t, 0)),
            ('E', (1, 0, 0)),
            ] 
        magn = 0.5

        lat, _ = LatLonLib.fromGeoDataFrameToLatLon(LatLonLib.NANTES)

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(2.0 * magn * 8.26, 1.1 * magn * 8.26))
        fig.suptitle(f'''Direct Solar Irradiation, lat={lat:+.1f}°''', size=20)
        MINS = range(0, 60, 30)
        for label, normal in PAIRS:
            X, Y = [], []
            for month in range(1, 13):
                for day in range(1, 28, 10):
                    _Y = []
                    for hour in range(4, 22):
                        for minutes in MINS:
                            dt = datetime(2021, month, day, hour, minutes)
                            _Y.append(DirectSolarIrradianceLib.noMaskDI(normal, dt))

                    dayInYear = dt.timetuple().tm_yday
                    X.append(dayInYear)
                    Y.append(sum(_Y) / (1e3 * len(MINS)))

            linestyle = '-' if (1 == len(label)) else ':'
            ax.plot(X, Y, linestyle=linestyle, linewidth=1.5,
                       label=f'{label} {10 * sum(Y) / 1e3:.2f} MWh/m$^2$')
        ax.set_xlabel('Day of year')
        ax.set_ylabel('Daily irradiation [kWh.m$^{-2}$]')

        for month, label in [(3, 'Spring equinox'), (6, 'Summer solstice'), (9, 'Autumn equinox')]:
            x0 = int(datetime(2021, month, 21).strftime('%j'))
            ax.axvline(x=x0, linestyle=':', color='blue', linewidth=1)
            ax.text(x0 + 5, 0.4, label, fontsize=10, rotation=90, color='blue')

        ax.set_ylim([0, 5.2])
        # ax.grid(color='gray', linestyle=':', linewidth=1)
        ax.legend(loc='upper left', framealpha=0.5, ncol=4, fontsize=8)

        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, format='pdf', bbox_inches='tight')
        plt.close(fig)

    @staticmethod
    def plotAtkinsonAnnuel(ofile=None):
        t = 1.0 / sqrt(2.0)
        PAIRS1 = [
            ('Roof', (0, 0, 1)),
            ('N', (0, 1, 0)),
            ('W', (-1, 0, 0)),
            ('S', (0, -1, 0)),
            ('E', (1, 0, 0)),
            ] 
        PAIRS2 = [
            ('Roof', (0, 0, 1)),
            ('NE', (t, t, 0)),
            ('NW', (-t, t, 0)),
            ('SW', (-t, -t, 0)),
            ('SE', (t, -t, 0)),
            ] 
        magn = 0.5

        lat, _ = LatLonLib.fromGeoDataFrameToLatLon(LatLonLib.NANTES)

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(2.0 * magn * 8.26, 1.1 * magn * 8.26))
        fig.suptitle(f'''Direct Solar Irradiation, lat={lat:+.1f}°''', size=20)
        MINS = range(0, 60, 30)
        for label, pairs in [('Roof+N+E+S+W', PAIRS1), ('Roof+NE+SE+SW+NW', PAIRS2)]:
            X, Y = [], []
            for month in range(1, 13):
                for day in range(1, 28, 10):
                    _Y = {0: [], 1: [], 2: [], 3: [], 4: []}
                    for hour in range(4, 22):
                        for minutes in MINS:
                            for i, (_, normal) in enumerate(pairs):
                                dt = datetime(2021, month, day, hour, minutes)
                                _Y[i].append(DirectSolarIrradianceLib.noMaskDI(normal, dt))

                    dayInYear = dt.timetuple().tm_yday
                    X.append(dayInYear)
                    acc = sum([sum(_Y[i]) for i in range(5)])
                    Y.append(acc / (1e3 * len(MINS)))

            linestyle = '-' if (1 == len(label)) else ':'
            ax.plot(X, Y, linestyle=linestyle, linewidth=1.5,
                       label=f'{label} {10 * sum(Y) / 1e3:.2f} MWh/m$^3$')
        ax.set_xlabel('Day of year')
        ax.set_ylabel('Daily irradiation [kWh]')

        for month, label in [(3, 'Spring equinox'), (6, 'Summer solstice'), (9, 'Autumn equinox')]:
            x0 = int(datetime(2021, month, 21).strftime('%j'))
            ax.axvline(x=x0, linestyle=':', color='blue', linewidth=1)
            ax.text(x0 + 5, 6, label, fontsize=10, rotation=90, color='blue')

        ax.set_ylim([5, 20])
        # ax.grid(color='gray', linestyle=':', linewidth=1)
        ax.legend(loc='upper left', framealpha=0.5, ncol=4, fontsize=8)

        if ofile is None:
            plt.show()
        else:
            plt.savefig(ofile, format='pdf', bbox_inches='tight')
        plt.close(fig)

'''
DirectSolarIrradianceLib.plot('South orientation', (0, -1, 0), '/tmp/south.pdf')
DirectSolarIrradianceLib.plot('East orientation', (1, 0, 0), '/tmp/east.pdf')
DirectSolarIrradianceLib.plot('West orientation', (-1, 0, 0), '/tmp/west.pdf')
DirectSolarIrradianceLib.plot('Roof', (0, 0, 1), '/tmp/roof.pdf')

DirectSolarIrradianceLib.plotFelixMarboutin('/tmp/felix_marboutin.pdf')
DirectSolarIrradianceLib.plotFelixMarboutinAnnuel('/tmp/felix_marboutin_annuel.pdf')
DirectSolarIrradianceLib.plotAtkinsonAnnuel('/tmp/atkinson_annuel.pdf')
'''



'''
Created on 19 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from shapely.geometry import LineString, Point

from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class DiameterLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def diameter(obj):
        if not GeomLib.isAShapelyGeometry(obj):
            raise IllegalArgumentTypeException(obj, 'Shapely geometry')

        vertices = [Point(c) for c in obj.convex_hull.exterior.coords]

        maxDist, pairOfVertices = -1.0, None
        for v1 in vertices:
            for v2 in vertices:
                currDist = v1.distance(v2)
                if maxDist < currDist:
                    maxDist = currDist
                    pairOfVertices = [v1, v2]
        
        if pairOfVertices is None:
            return [None, None, None]
        
        segment = LineString(pairOfVertices)
        return [segment, maxDist, GeomLib.getLineStringOrientation(segment)]



'''
Created on 1 oct. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import matplotlib.pyplot as plt
from numpy import histogram
from pandas import DataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class PopulationPyramid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, male_df, female_df, commonFieldname, nClusters='auto',
                 outputFile=None, title=None):
        '''
        Constructor
        '''
        for _gdf in (male_df, female_df):
            if not isinstance(_gdf, DataFrame):
                raise IllegalArgumentTypeException(_gdf, 'DataFrame')
            if commonFieldname not in _gdf:
                raise Exception('%s is not a relevant field name!' % commonFieldname)
        self.males = male_df[commonFieldname] 
        self.females = female_df[commonFieldname] 
        self.nClusters = nClusters
        self.outputFile = outputFile
        self.title = title

    def run(self):
        m_hist, m_bin_edges = histogram(self.males, bins=self.nClusters)
        f_hist, _ = histogram(self.females, bins=len(m_hist))
        xmin, xmax = min(min(m_hist), min(f_hist)), 1.01 * max(max(m_hist), max(f_hist))

        fig, axes = plt.subplots(ncols=2, sharey=True, figsize=(9, 6))
        if self.title is None:
            plt.suptitle('Population Pyramid', fontsize=20, ha='center')
        else:
            plt.suptitle(self.title, fontsize=20, ha='center')
        # y = [(m_bin_edges[i + 1] + m_bin_edges[i]) / 2.0 for i in range(len(m_hist))]
        yLbls = ['%.1f-%.1f' % (m_bin_edges[i], m_bin_edges[i + 1]) for i in range(len(m_hist))]

        axes[0].barh(yLbls, m_hist, align='center', color='darkgrey')
        axes[0].set(title='Males')
        axes[0].set_xlim([xmin, xmax])
        axes[0].set(yticks=yLbls, yticklabels=yLbls)
        axes[0].invert_xaxis()
        axes[0].grid()

        axes[1].barh(yLbls, f_hist, align='center', color='lightgrey')
        axes[1].set(title='Females')
        axes[1].set_xlim([xmin, xmax])
        axes[1].grid()

        if self.outputFile is None:
            fig.tight_layout()
            plt.show()
        else:
            plt.savefig(self.outputFile, format='pdf', bbox_inches='tight')
        plt.close(fig)



'''
Created on 10 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timezone

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import MultiPoint, MultiPolygon, Point
from t4gpd.commons.DatetimeLib import DatetimeLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.ShadowLib import ShadowLib
from t4gpd.commons.sun.SunLib import SunLib
from t4gpd.sun.AbstractHardShadow import AbstractHardShadow


class STTreeHardShadow2(AbstractHardShadow):
    '''
    classdocs
    '''

    def __init__(self, treesGdf, datetimes, treeHeightFieldname, treeCrownHeightFieldname,
                 treeUpperCrownRadiusFieldname, treeLowerCrownRadiusFieldname,
                 altitudeOfShadowPlane=0, aggregate=False, tz=timezone.utc, model='pysolar',
                 npoints=32):
        '''
        Constructor
        '''
        if not isinstance(treesGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(treesGdf, 'GeoDataFrame')
        self.gdf = treesGdf
        self.crs = treesGdf.crs

        if treeHeightFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (treeHeightFieldname))
        self.treeHeightFieldname = treeHeightFieldname

        if treeCrownHeightFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (treeCrownHeightFieldname))
        self.treeCrownHeightFieldname = treeCrownHeightFieldname

        if treeUpperCrownRadiusFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (treeUpperCrownRadiusFieldname))
        self.treeUpperCrownRadiusFieldname = treeUpperCrownRadiusFieldname

        if treeLowerCrownRadiusFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (treeLowerCrownRadiusFieldname))
        self.treeLowerCrownRadiusFieldname = treeLowerCrownRadiusFieldname

        self.altitudeOfShadowPlane = altitudeOfShadowPlane
        self.aggregate = aggregate
        sunModel = SunLib(gdf=treesGdf, model=model)

        self.sunPositions = DatetimeLib.fromDatetimesDictToListOfSunPositions(
            datetimes, sunModel, tz)

        if not (0 == (npoints % 4)):
            raise IllegalArgumentTypeException(npoints, 'multiple of 4')
        self.npoints = npoints

    def _auxiliary(self, row, radDir, solarAlti, solarAzim):
        treeGeom = row.geometry
        treeHeight = row[self.treeHeightFieldname]
        treeCrownHeight = row[self.treeCrownHeightFieldname]
        treeUpperCrownRadius = row[self.treeUpperCrownRadiusFieldname]
        treeLowerCrownRadius = row[self.treeLowerCrownRadiusFieldname]

        if isinstance(treeGeom, Point):
            _shadow = ShadowLib.projectConicalTreeOntoShadowPlane(
                treeGeom, treeHeight, treeCrownHeight, treeUpperCrownRadius,
                treeLowerCrownRadius, None, radDir, solarAlti, solarAzim,
                self.altitudeOfShadowPlane, self.npoints)
            return _shadow

        elif isinstance(treeGeom, MultiPoint):
            _shadows = []
            for g in treeGeom.geoms:
                _shadow = ShadowLib.projectConicalTreeOntoShadowPlane(
                    g, treeHeight, treeCrownHeight, treeUpperCrownRadius,
                    treeLowerCrownRadius, None, radDir, solarAlti, solarAzim,
                    self.altitudeOfShadowPlane, self.npoints)
                if _shadow is not None:
                    _shadows.append(_shadow)
            return None if (0 == len(_shadows)) else MultiPolygon(_shadows)
